import re
from re import Pattern
from time import sleep
from typing import TYPE_CHECKING, Optional

import tmt.log
import tmt.utils
from tmt.checks import Check, CheckPlugin, _RawCheck, provides_check
from tmt.container import container, field
from tmt.result import CheckResult, ResultOutcome
from tmt.utils import Command, Path, ShellScript

if TYPE_CHECKING:
    from tmt.steps.execute import TestInvocation
    from tmt.steps.provision import Guest


COREDUMP_LAST_DUMP_FILENAME = "coredump-latest"

# Can be set in /etc/coredumpct.conf.d/
# See `man coredump.conf`
COREDUMP_CONFIG = """[Coredump]
Storage=external
Compress=yes
"""


@container
class CoredumpCheck(Check):
    """Configuration for the coredump check."""

    # Patterns to ignore in crash reports
    ignore_pattern: list[Pattern[str]] = field(
        default_factory=list,
        help="""
             Optional list of regular expressions to ignore in crash reports.
             If a crash report matches any of these patterns, it will be ignored
             and not cause a failure. Any other crashes will still cause the test
             to fail. If no patterns are specified, any crash will cause a failure.
             """,
        metavar="PATTERN",
        normalize=tmt.utils.normalize_pattern_list,
        exporter=lambda patterns: [pattern.pattern for pattern in patterns],
        serialize=lambda patterns: [pattern.pattern for pattern in patterns],
        unserialize=lambda serialized: [re.compile(pattern) for pattern in serialized],
    )

    # Internal flag to track if we can run coredumpctl on the host
    is_available: bool = True
    is_availability_reason: Optional[str] = None

    # Path to the file storing information about coredumps before test execution
    # Default is current directory, will be properly set in _save_existing_coredumps
    coredump_last_dumps_filepath: Path = field(default_factory=Path, internal=True)

    def to_spec(self) -> _RawCheck:
        """Convert to raw specification."""
        spec = super().to_spec()

        spec["ignore-pattern"] = [  # type: ignore[reportGeneralTypeIssues,typeddict-unknown-key,unused-ignore]
            pattern.pattern for pattern in self.ignore_pattern
        ]

        return spec

    def to_minimal_spec(self) -> _RawCheck:
        """Convert to minimal raw specification."""
        return self.to_spec()

    def _configure_coredump(self, guest: "Guest", logger: tmt.log.Logger) -> None:
        """
        Try configure coredump storage.

        Non-privileged users might not have permission to change the config,
        while being able to use coredumpctl
        """
        try:
            guest.execute(
                ShellScript("mkdir -p /etc/systemd/coredump.conf.d")
                & ShellScript(
                    f"echo '{COREDUMP_CONFIG}' > /etc/systemd/coredump.conf.d/50-tmt.conf"
                )
            )
            return
        except tmt.utils.RunError:
            logger.debug("Unable to configure coredump directly, trying with sudo")

        # If failed and not root, try with sudo
        if guest.facts.is_superuser is False:
            try:
                guest.execute(
                    ShellScript("sudo mkdir -p /etc/systemd/coredump.conf.d")
                    & ShellScript(
                        f"echo '{COREDUMP_CONFIG}' | sudo tee "
                        "/etc/systemd/coredump.conf.d/50-tmt.conf > /dev/null"
                    )
                )
                logger.debug("Configured coredump with sudo")
                return
            except tmt.utils.RunError:
                logger.debug(
                    "Unable to configure coredump even with sudo, continuing with default settings"
                )
        else:
            logger.debug("Unable to configure coredump, continuing with default settings")

    def _has_coredumpctl(self, guest: "Guest", logger: tmt.log.Logger) -> bool:
        """Check if coredumpctl command is available."""
        try:
            guest.execute(ShellScript("coredumpctl --version"), silent=True)
            return True
        except tmt.utils.RunError:
            logger.debug("coredumpctl command not found")
            return False

    def _has_active_socket(self, guest: "Guest", logger: tmt.log.Logger) -> bool:
        """Check if systemd-coredump.socket is active or can be activated."""
        try:
            # Try activating the socket if it's not already active
            guest.execute(
                ShellScript(
                    "systemctl is-active systemd-coredump.socket || "
                    "systemctl start systemd-coredump.socket"
                ),
                silent=True,
            )
            return True
        except tmt.utils.RunError:
            # If the user isn't root, try with sudo
            if guest.facts.is_superuser is False:
                try:
                    guest.execute(
                        ShellScript(
                            "sudo systemctl is-active systemd-coredump.socket || "
                            "sudo systemctl start systemd-coredump.socket"
                        ),
                        silent=True,
                    )
                    logger.debug("systemd-coredump.socket activated with sudo")
                    return True
                except tmt.utils.RunError:
                    logger.debug(
                        "Unable to access or start systemd-coredump.socket, even with sudo"
                    )
                    return False
            else:
                logger.debug("Unable to access or start systemd-coredump.socket")
                return False

    def _has_required_permissions(self, guest: "Guest", logger: tmt.log.Logger) -> bool:
        """
        Check if we have sufficient permissions to access coredump data.

        Note: coredumpctl returns 1 when no dumps are found, which is a valid
        case that indicates we have proper permissions.
        """
        # Try without sudo first
        try:
            # Permission errors shouldn't return 1, while "no dumps found" does.
            # See https://github.com/systemd/systemd/blob/main/src/basic/errno-util.h
            guest.execute(ShellScript("coredumpctl info &>/dev/null || [ $? -eq 1 ]"), silent=True)
            logger.debug("Has permission to access coredump data")
            return True
        except tmt.utils.RunError:
            # If failed, try with sudo if the user is not already root
            if guest.facts.is_superuser is False:
                try:
                    guest.execute(
                        ShellScript("sudo coredumpctl info &>/dev/null || [ $? -eq 1 ]"),
                        silent=True,
                    )
                    logger.debug("Access to coredump data requires sudo, which is available")
                    return True
                except tmt.utils.RunError:
                    logger.debug("Cannot access coredump data even with sudo - permission issues")
                    return False
            else:
                logger.debug("Cannot access coredump data - permission issues")
                return False

    def _check_coredump_available(
        self, guest: "Guest", logger: tmt.log.Logger
    ) -> tuple[bool, str]:
        """
        Check if coredump functionality is available and usable.

        Checks for:
        1. systemd availability through guest facts
        2. coredumpctl command exists
        3. systemd-coredump.socket is active or can be activated
        4. Has sufficient permissions to access coredump data

        :returns: A tuple of (available, reason) where available is True if coredump
                 is available and usable, and reason is a string explaining why it's not
                 available if applicable.
        """
        # We need systemd for coredump functionality
        if not guest.facts.has_systemd:
            reason = "systemd not available"
            logger.debug(f"{reason}, skipping coredump check")
            return False, reason

        # Check for coredumpctl command
        if not self._has_coredumpctl(guest, logger):
            reason = "coredumpctl command not available"
            return False, reason

        # Check for systemd-coredump.socket
        if not self._has_active_socket(guest, logger):
            reason = "systemd-coredump.socket not active and could not be activated"
            return False, reason

        # Check for permissions
        if not self._has_required_permissions(guest, logger):
            reason = "insufficient permissions to access coredump data"
            return False, reason

        return True, ""

    def _wait_for_coredump_processes(self, guest: "Guest", logger: tmt.log.Logger) -> None:
        """
        Wait for systemd-coredump processes to finish processing dumps.

        Uses progressive waiting with a timeout to avoid getting stuck.
        """
        need_sudo = guest.facts.is_superuser is False
        sudo_prefix = "sudo " if need_sudo else ""

        total_wait = 0
        max_wait = 60  # Total maximum wait time in seconds
        wait_time = 1  # Start with 1 second, will increase progressively

        logger.debug("Checking if systemd-coredump processes are running")
        while total_wait < max_wait:
            try:
                # Check if any systemd-coredump processes are running
                cmd = f"{sudo_prefix}pgrep systemd-coredump || true"
                result = guest.execute(ShellScript(cmd), silent=True)

                # If no processes found, we're good to go
                if not result.stdout:
                    logger.debug("No systemd-coredump processes found")
                    return

                # If processes are still running, wait and try again
                process_count = len(result.stdout.splitlines())
                logger.debug(
                    f"Found {process_count} systemd-coredump processes, waiting {wait_time}s"
                )
                sleep(wait_time)

                # Increase wait time progressively (1, 2, 3, 4...)
                total_wait += wait_time
                wait_time = min(wait_time + 1, 5)  # Cap at 5 seconds per wait

            except tmt.utils.RunError:
                # If we can't check, just return
                logger.debug("Unable to check for systemd-coredump processes")
                return

        logger.debug(f"Timed out after {total_wait}s waiting for systemd-coredump processes")

    def _check_for_crashes(
        self,
        guest: "Guest",
        logger: tmt.log.Logger,
        check_files_path: Path,
        previous_dumps_file: Path,
    ) -> bool:
        """
        Check if any new crashes have been detected since the test started.

        Uses the timestamp of the latest coredump before the test to identify
        new coredumps created during test execution.
        """
        # Determine if we need sudo
        need_sudo = guest.facts.is_superuser is False
        sudo_prefix = "sudo " if need_sudo else ""

        # Make sure dumps are processed
        self._wait_for_coredump_processes(guest, logger)

        try:
            # Get the timestamp of the latest coredump before the test
            latest_timestamp = ""
            try:
                previous_output = guest.execute(Command("cat", previous_dumps_file)).stdout
                if previous_output:
                    # The output should contain just one line for the latest coredump
                    # Extract the timestamp (first two fields) if we have data
                    fields = previous_output.strip().split()
                    if len(fields) >= 2:
                        # Combine the date and time fields as the timestamp
                        latest_timestamp = f"{fields[0]} {fields[1]}"
                        logger.debug(
                            f"Latest coredump before test had timestamp: {latest_timestamp}"
                        )
            except tmt.utils.RunError:
                logger.debug("Unable to read previous dumps file, assuming no prior coredumps")

            # Get list of coredumps newer than the latest one before the test
            # Use --all to check coredumps from all users, not just current user
            if latest_timestamp:
                since_param = f'--since="{latest_timestamp}"'
                cmd = f"{sudo_prefix}coredumpctl list --all --no-legend --no-pager {since_param}"
                logger.debug(f"Checking for coredumps newer than: {latest_timestamp}")
            else:
                # If we have no prior timestamp, get all coredumps
                cmd = f"{sudo_prefix}coredumpctl list --all --no-legend --no-pager"
                logger.debug("No prior coredumps found, checking all available coredumps")

            output = guest.execute(ShellScript(cmd)).stdout

            if not output:
                return False

            # Process each crash entry
            has_new_crashes = False
            for line in output.splitlines():
                fields = line.split()
                if len(fields) < 10:  # Ensure we have enough fields
                    logger.debug(f"Skipping malformed coredump line: {line}")
                    continue

                pid = fields[4]
                sig = fields[7]
                corefile = fields[8]
                exe = fields[9].replace("/", "_")

                # We're now using the timestamp to filter,
                # so all dumps in this list are already post-test

                # Get detailed info for this crash (use --all to ensure we can access it)
                cmd = f"{sudo_prefix}coredumpctl info --all --no-pager {pid}"
                crash_info = guest.execute(ShellScript(cmd)).stdout
                if not crash_info:
                    logger.debug(f"No crash info available for PID {pid}")
                    continue

                # Skip if this crash matches any ignore pattern
                if self.ignore_pattern:
                    matching_pattern = next(
                        (pattern for pattern in self.ignore_pattern if pattern.search(crash_info)),
                        None,
                    )
                    if matching_pattern:
                        logger.info(
                            f"Ignoring crash in PID {pid} ({exe}) due to pattern match: '{matching_pattern.pattern}'"  # noqa: E501
                        )
                        # Log full crash info at debug level
                        logger.debug(f"Full ignored crash info: {crash_info}")
                        continue

                # Save the crash info (with --all to ensure we can access it)
                info_filepath = check_files_path / f"dump.{exe}_{sig}_{pid}.txt"
                guest.execute(
                    ShellScript(
                        f"sh -c {sudo_prefix}coredumpctl info --all --no-pager {pid} > {info_filepath!s}"  # noqa: E501
                    )
                )
                logger.debug(f"Saved crash info to {info_filepath}")

                # Try to save the coredump if available
                if corefile not in ("none", "missing"):
                    dump_filename = f"dump.{exe}_{sig}_{pid}.core"
                    dump_path = check_files_path / dump_filename
                    try:
                        guest.execute(
                            ShellScript(
                                f"{sudo_prefix}coredumpctl dump --all --no-pager -o {dump_path!s} {pid}"  # noqa: E501
                            )
                        )
                        logger.debug(f"Saved coredump to {dump_path}")
                    except tmt.utils.RunError as exc:
                        logger.debug(f"Failed to save coredump for PID {pid}: {exc}")
                else:
                    logger.debug(f"Skipping dump for PID {pid}, corefile status: {corefile}")

                # This is a new, non-ignored crash
                has_new_crashes = True

            return has_new_crashes

        except tmt.utils.RunError as exc:
            logger.debug(f"Failed to check for crashes: {exc}")
            return False

    def _check_coredump(
        self, invocation: "TestInvocation", logger: tmt.log.Logger
    ) -> tuple[ResultOutcome, list[Path]]:
        """
        Check coredump status and return appropriate result.

        :returns: A tuple of (outcome, log_files) where log_files is a list of
                 paths to files with coredump information.
        """
        log_files: list[Path] = []

        # Check for crashes by comparing with the saved file of previous dumps
        has_crashes = self._check_for_crashes(
            invocation.guest,
            logger,
            invocation.check_files_path,
            self.coredump_last_dumps_filepath,
        )

        # Get list of generated files
        try:
            files_output = invocation.guest.execute(
                Command("find", invocation.check_files_path, "-type", "f")
            ).stdout

            if files_output:
                for file_path in files_output.splitlines():
                    path = Path(file_path)
                    if path.exists() and invocation.phase.step.workdir:
                        rel_path = path.relative_to(invocation.phase.step.workdir)
                        log_files.append(rel_path)
        except tmt.utils.RunError:
            logger.debug("Failed to list coredump files")

        if has_crashes:
            return ResultOutcome.FAIL, log_files

        return ResultOutcome.PASS, log_files

    def _save_existing_coredumps(self, invocation: "TestInvocation") -> bool:
        """
        Save information about the latest coredump before the test.

        This allows us to only check for newer coredumps after the test.
        """
        self.coredump_last_dumps_filepath = (
            invocation.check_files_path / COREDUMP_LAST_DUMP_FILENAME
        )

        try:
            # Create directory for check files
            invocation.guest.execute(ShellScript(f"mkdir -p {invocation.check_files_path!s}"))

            # Determine if sudo is needed
            need_sudo = invocation.guest.facts.is_superuser is False
            sudo_prefix = "sudo " if need_sudo else ""

            # Get only the latest coredump before the test using -1 flag
            # and save it to a file. If there are no dumps, create an empty file.
            # Use --all to check coredumps from all users, not just current user
            invocation.guest.execute(
                ShellScript(
                    f"({sudo_prefix}coredumpctl list -1 --all --no-legend --no-pager || true) > "
                    f"{self.coredump_last_dumps_filepath!s}"
                )
            )
            return True
        except tmt.utils.RunError:
            return False


# TODO: Enable hints when PR #3498 is merged
# @provides_check(
#     "coredump",
#     hints={  # type: ignore[call-arg]
#         "not-available": """
#             Coredump detection was skipped because coredumpctl is not available or has
#             insufficient privileges.


#             The coredump check requires systemd-coredump to be installed and the
#             systemd-coredump.socket to be active. Additionally, the user must have
#             sufficient permissions to access coredump data.
#             """
#     },
# )
@provides_check("coredump")
class Coredump(CheckPlugin[CoredumpCheck]):
    """
    Check for system crashes using coredump.

    The check monitors for any crashes caught by systemd-coredump during test
    execution. This includes segmentation faults and other crashes that produce
    core dumps. By default, any crash will cause the test to fail.

    Example config with optional ignore patterns:

    .. code-block:: yaml

        check:
          - how: coredump
            ignore-pattern:
              - 'Process.*\\(sleep\\).*dumped core'  # Ignore sleep crashes
              - 'Package: ddcutil/2.1.2-2.fc41'      # Ignore dumps of a specific package

    The patterns are matched against the full coredumpctl info output, which includes
    fields like Process, Command Line, Signal, etc. You can use 'coredumpctl info'
    to see the available fields and their format.

    Common pattern examples:

    - Ignore all crashes from a specific process:
      `'Process.*\\(specific-process\\)'`

    - Ignore crashes with a specific signal:
      `'Signal: .*\\(SIGSEGV\\)'`

    - Ignore crashes from a specific package:
      `'Package: package-name/.*'`

    - Ignore crashes during a specific command:
      `'Command Line: .*specific-command-pattern.*'`

    .. versionadded:: 1.46
    """

    @staticmethod
    def _get_not_available_message(reason: Optional[str] = None) -> str:
        """Return a message explaining why coredump is not available."""
        if reason:
            base_message = f"Coredump detection was skipped: {reason}."
        else:
            base_message = (
                "Coredump detection was skipped because coredumpctl is not available or has "
                "insufficient privileges."
            )

        return (
            f"{base_message}\n\n"
            "The coredump check requires systemd-coredump to be installed and the "
            "systemd-coredump.socket to be active. Additionally, the user must have "
            "sufficient permissions to access coredump data."
        )

    _check_class = CoredumpCheck

    @classmethod
    def before_test(
        cls,
        *,
        check: "CoredumpCheck",
        invocation: "TestInvocation",
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        """Check for crashes before the test starts."""

        # TODO: Uncomment when PR #3498 is merged
        # from tmt.utils.hints import get_hints

        available, reason = check._check_coredump_available(invocation.guest, logger)
        if not available:
            logger.debug(f"coredump not available: {reason}, skipping..")
            check.is_availability_reason = reason
            check.is_available = False
            # Use our temporary message until PR #3498 is merged
            return [
                CheckResult(
                    name="coredump",
                    result=ResultOutcome.SKIP,
                    note=[cls._get_not_available_message(reason)],
                )
            ]

        # Initialize outcome to ERROR in case we can't save the current coredumps list
        outcome = ResultOutcome.ERROR
        log_files: list[Path] = []

        if check._save_existing_coredumps(invocation):
            check._configure_coredump(invocation.guest, logger)
            outcome, log_files = check._check_coredump(invocation, logger)

        return [CheckResult(name="coredump", result=outcome, log=log_files)]

    @classmethod
    def after_test(
        cls,
        *,
        check: "CoredumpCheck",
        invocation: "TestInvocation",
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        """Check for crashes after the test finishes."""
        # TODO: Uncomment when PR #3498 is merged
        # from tmt.utils.hints import get_hints

        # Skip if coredumpctl is not available, as detected in before-test
        if not check.is_available:
            return [
                CheckResult(
                    name="coredump",
                    result=ResultOutcome.SKIP,
                    note=[cls._get_not_available_message(check.is_availability_reason)],
                )
            ]

        if not invocation.is_guest_healthy:
            return [
                CheckResult(
                    name="coredump",
                    result=ResultOutcome.SKIP,
                    note=["Coredump check skipped because the guest is not healthy."],
                )
            ]

        outcome, log_files = check._check_coredump(invocation, logger)
        return [CheckResult(name="coredump", result=outcome, log=log_files)]



import tempfile
from typing import Optional, Union

import requests

import tmt
import tmt.base
import tmt.log
import tmt.steps
import tmt.steps.prepare
import tmt.steps.provision
import tmt.utils
from tmt.container import container, field
from tmt.result import PhaseResult
from tmt.steps.provision import (
    ANSIBLE_COLLECTION_PLAYBOOK_PATTERN,
    AnsibleApplicable,
    AnsibleCollectionPlaybook,
    Guest,
)
from tmt.utils import (
    DEFAULT_RETRIABLE_HTTP_CODES,
    ENVFILE_RETRY_SESSION_RETRIES,
    Path,
    PrepareError,
    normalize_string_list,
    retry_session,
)


class _RawAnsibleStepData(tmt.steps._RawStepData, total=False):
    playbook: Union[str, list[str]]
    playbooks: list[str]


@container
class PrepareAnsibleData(tmt.steps.prepare.PrepareStepData):
    playbook: list[str] = field(
        default_factory=list,
        option=('-p', '--playbook'),
        multiple=True,
        metavar='PATH|URL',
        help="""
             Path or URL of an ansible playbook to run.
             The playbook path must be relative to the metadata tree root.
             """,
        normalize=tmt.utils.normalize_string_list,
    )

    extra_args: Optional[str] = field(
        default=None,
        option='--extra-args',
        metavar='ANSIBLE-PLAYBOOK-OPTIONS',
        help='Additional CLI options for ``ansible-playbook``.',
        help_example_values=['-vvv'],
    )

    # ignore[override]: method violates a liskov substitution principle,
    # but only apparently.  Thanks to how tmt initializes module, we can
    # safely assume PrepareAnsibleData.pre_normalization() would be
    # called with source data matching _RawAnsibleStepData.
    @classmethod
    def pre_normalization(  # type: ignore[override]
        cls, raw_data: _RawAnsibleStepData, logger: tmt.log.Logger
    ) -> None:
        super().pre_normalization(raw_data, logger)

        # Perform `playbook` normalization here, so we could merge `playbooks` to it.
        playbook = normalize_string_list('playbook', raw_data.pop('playbook', []), logger)
        playbooks = normalize_string_list('playbook', raw_data.pop('playbooks', []), logger)

        raw_data['playbook'] = [*playbook, *playbooks]


@tmt.steps.provides_method('ansible')
class PrepareAnsible(tmt.steps.prepare.PreparePlugin[PrepareAnsibleData]):
    #
    # This plugin docstring has been reviewed and updated to follow
    # our documentation best practices. When changing it, please make
    # sure new changes are following them as well.
    #
    # https://tmt.readthedocs.io/en/stable/contribute.html#docs
    #
    """
    Run Ansible playbooks against the guest, by running
    ``ansible-playbook`` for all given playbooks.

    .. note::

       The plugin requires a working Ansible to be available on the
       test runner.

    .. warning::

        When specifying playbooks with paths:

        * If a metadata tree root exists, all paths must be relative to
          the metadata tree root.
        * If the metadata tree root does not exist,
          all paths must be relative to the current working directory.

    .. warning::

       The plugin may be a subject of various limitations, imposed by
       Ansible itself:

       * Ansible 2.17+ no longer supports Python 3.6 and older. Guests
         where Python 3.7+ is not available cannot be prepared with the
         ``ansible`` plugin. This has been observed when Fedora Rawhide
         runner is used with CentOS 7 or CentOS Stream 8 guests. Possible
         workarounds: downgrade Ansible tmt uses, or install Python 3.7+
         before using ``ansible`` plugin from an alternative repository
         or local build.

    Run a single playbook on the guest:

    .. code-block:: yaml

        prepare:
            how: ansible
            playbook: ansible/packages.yml

    .. code-block:: shell

        prepare --how ansible --playbook ansible/packages.yml

    Run multiple playbooks in one phase, with extra arguments for
    ``ansible-playbook``:

    .. code-block:: yaml

        prepare:
            how: ansible
            playbook:
              - one.yml
              - two.yml
            extra-args: '-vvv'

    .. code-block:: shell

        prepare --how ansible --playbook one.yml --playbook two.yml --extra-args '-vvv'

    Remote playbooks - provided as URLs starting with ``http://`` or
    ``https://``, local playbooks - optionally starting with a
    ``file://`` schema, and playbooks bundled with collections can be
    referenced as well as local ones, and all kinds can be intermixed:

    .. code-block:: yaml

        prepare:
            how: ansible
            playbook:
              - https://foo.bar/one.yml
              - two.yml
              - file://three.yml
              - ansible_galaxy_namespace.cool_collection.four

    .. code-block:: shell

        prepare --how ansible --playbook https://foo.bar/two.yml \\
                              --playbook two.yml \\
                              --playbook file://three.yml \\
                              --playbook ansible_galaxy_namespace.cool_collection.four
    """

    _data_class = PrepareAnsibleData

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Prepare the guests
        """

        results = super().go(guest=guest, environment=environment, logger=logger)

        # Apply each playbook on the guest
        for _playbook in self.data.playbook:
            logger.info('playbook', _playbook, 'green')

            lowercased_playbook = _playbook.lower()

            def normalize_remote_playbook(raw_playbook: str) -> tuple[Path, AnsibleApplicable]:
                assert self.step.workdir is not None  # narrow type
                root_path = self.step.workdir

                try:
                    with retry_session(
                        retries=ENVFILE_RETRY_SESSION_RETRIES,
                        status_forcelist=DEFAULT_RETRIABLE_HTTP_CODES,
                    ) as session:
                        response = session.get(raw_playbook)

                    if not response.ok:
                        raise PrepareError(f"Failed to fetch remote playbook '{raw_playbook}'.")

                except requests.RequestException as exc:
                    raise PrepareError(
                        f"Failed to fetch remote playbook '{raw_playbook}'."
                    ) from exc

                with tempfile.NamedTemporaryFile(
                    mode='w+b',
                    prefix='playbook-',
                    suffix='.yml',
                    dir=root_path,
                    delete=False,
                ) as file:
                    file.write(response.content)
                    file.flush()

                    return root_path, Path(file.name).relative_to(root_path)

            def normalize_local_playbook(raw_playbook: str) -> tuple[Path, AnsibleApplicable]:
                if raw_playbook.startswith('file://'):
                    return self.step.plan.anchor_path, Path(raw_playbook[7:])

                return self.step.plan.anchor_path, Path(raw_playbook)

            def normalize_collection_playbook(raw_playbook: str) -> tuple[Path, AnsibleApplicable]:
                return self.step.plan.anchor_path, AnsibleCollectionPlaybook(raw_playbook)

            if lowercased_playbook.startswith(('http://', 'https://')):
                playbook_root, playbook = normalize_remote_playbook(lowercased_playbook)

            elif lowercased_playbook.startswith('file://'):
                playbook_root, playbook = normalize_local_playbook(lowercased_playbook)

            elif ANSIBLE_COLLECTION_PLAYBOOK_PATTERN.match(lowercased_playbook):
                playbook_root, playbook = normalize_collection_playbook(lowercased_playbook)

            else:
                playbook_root, playbook = normalize_local_playbook(lowercased_playbook)

            guest.ansible(
                playbook,
                playbook_root=playbook_root,
                extra_args=self.data.extra_args,
            )

        return results

    def essential_requires(self) -> list[tmt.base.Dependency]:
        """
        Collect all essential requirements of the plugin.

        Essential requirements of a plugin are necessary for the plugin to
        perform its basic functionality.

        :returns: a list of requirements.
        """

        return tmt.steps.provision.essential_ansible_requires()



import datetime
import re
import threading
import time
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional

import tmt.log
import tmt.steps.provision
import tmt.steps.provision.artemis
import tmt.steps.provision.connect
import tmt.steps.provision.local
import tmt.steps.provision.mrack
import tmt.utils
from tmt.checks import Check, CheckPlugin, provides_check
from tmt.container import container, field
from tmt.result import CheckResult, ResultOutcome
from tmt.utils import Path, format_timestamp, render_run_exception_streams

if TYPE_CHECKING:
    from tmt.steps.execute import TestInvocation

PING_OUTPUT_PATTERN = re.compile(
    r'(?m)(?P<transmitted>\d+) packets transmitted, (?P<received>\d+) received'
)
SSH_PING_OUTPUT_PATTERN = re.compile(r'Ncat: Connected')

# TODO: do not use the list of classes, it's hard to maintain.
# Tracked in https://github.com/teemtee/tmt/issues/2739
PINGABLE_GUEST_CLASSES: tuple[type[tmt.steps.provision.Guest], ...] = (
    tmt.steps.provision.artemis.GuestArtemis,
    tmt.steps.provision.connect.GuestConnect,
    tmt.steps.provision.mrack.GuestBeaker,
    # TODO: is there a way to ping the VM instead of localhost?
    # Tracked in https://github.com/teemtee/tmt/issues/2738
    # tmt.steps.provision.testcloud.GuestTestcloud
)

SSH_PINGABLE_GUEST_CLASSES: tuple[type[tmt.steps.provision.Guest], ...] = (
    tmt.steps.provision.GuestSsh,
    tmt.steps.provision.local.GuestLocal,
)


REPORT_FILENAME = 'tmt-watchdog.txt'


def render_report_path(invocation: 'TestInvocation') -> Path:
    """
    Render path to a watchdog report file from necessary components
    """

    return invocation.check_files_path / REPORT_FILENAME


def report_progress(
    log: Path,
    check_name: str,
    report: Iterable[str],
    command_output: Optional[str] = None,
) -> None:
    """
    Add new report into a report file.

    :param log: path to the report file.
    :param report: iterable of report lines to add. Each line is emitted on its
        own line in the file.
    :param command_output: if set, the string is added to the report file once
        ``report`` lines are written into it.
    """

    timestamp = format_timestamp(datetime.datetime.now(datetime.timezone.utc))

    with open(log, mode='a') as f:
        f.write(f'# {check_name} reported at {timestamp}\n')

        for line in report:
            f.write(line)
            f.write('\n')

        if command_output:
            f.write('\n')
            f.write(command_output)

        f.write('\n')


@container
class GuestContext:
    """
    Per-guest watchdog context
    """

    #: Current number of failed watchdog checks.
    ping_failures: int = 0
    ssh_ping_failures: int = 0

    #: If set, contains a daemonized thread running the watchdog checks.
    thread: Optional[threading.Thread] = None

    #: As long as this field is set to ``True``, the watchdog will run its
    #: internal loop and run relevant checks. It is unset when terminating
    #: the watchdog check to notify the thread it's time to quit.
    keep_running: bool = True


@container
class WatchdogCheck(Check):
    interval: int = field(
        default=60,
        metavar='SECONDS',
        help='How often should the watchdog run.',
        normalize=tmt.utils.normalize_int,
    )

    reboot: bool = field(
        default=False,
        is_flag=True,
        help='If enabled, watchdog would reboot the guest after enough failed probes.',
    )

    ping: bool = field(
        default=False,
        is_flag=True,
        help="If enabled, watchdog would probe guest's responsiveness with ICMP packets.",
    )
    ping_packets: int = field(
        default=1,
        metavar='N',
        help='How many ICMP packates to send as one probe.',
        normalize=tmt.utils.normalize_int,
    )
    ping_threshold: int = field(
        default=10,
        metavar='N',
        help='How many failed ping probes before taking any further action.',
        normalize=tmt.utils.normalize_int,
    )

    ssh_ping: bool = field(
        default=False,
        is_flag=True,
        help="""
             If enabled, watchdog would probe guest's responsiveness by connecting
             to its SSH port.
             """,
    )
    ssh_ping_threshold: int = field(
        default=10,
        metavar='N',
        help='How many failed SSH connections before taking any further action.',
        normalize=tmt.utils.normalize_int,
    )

    def notify(self, invocation: 'TestInvocation', logger: tmt.log.Logger) -> None:
        """
        Notify invocation that hard reboot is required
        """

        if not self.reboot:
            return

        invocation.hard_reboot_requested = True
        invocation.terminate_process(logger=logger)

    def do_ping(
        self,
        invocation: 'TestInvocation',
        guest_context: GuestContext,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Perform a ping check
        """

        logger.debug('pinging', level=4)

        log = render_report_path(invocation)

        def _fail_parse_error(ping_output: str) -> None:
            """
            Handle unparsable ``ping`` output
            """

            logger.fail('failed to parse ping output')

            guest_context.ping_failures += 1

            report_progress(
                log,
                'ping',
                [
                    '# failed to parse ping output',
                    f'# failed {guest_context.ping_failures} of {self.ping_threshold} allowed',
                ],
                command_output=ping_output,
            )

        def _fail_lost_packets(ping_output: str, transmitted: int, received: int) -> None:
            """
            Handle missing response packets
            """

            logger.fail(f'not all packets returned: {transmitted=} {received=}')

            guest_context.ping_failures += 1

            report_progress(
                log,
                'ping',
                [
                    '# not all packets returned',
                    f'# failed {guest_context.ping_failures} of {self.ping_threshold} allowed',
                ],
                command_output=ping_output,
            )

        def _success(ping_output: str) -> None:
            """
            Handle successful response
            """

            logger.verbose('Received successful response to ping.', level=2)

            report = ['# successful response']

            if guest_context.ping_failures != 0:
                report.append(f'# replenished failure budget back to {self.ping_threshold}')

            guest_context.ping_failures = 0

            report_progress(log, 'ping', report, command_output=ping_output)

        def _handle_output(ping_output: str) -> None:
            """
            Process ``ping`` output and decide on its outcome
            """

            match = PING_OUTPUT_PATTERN.search(ping_output)

            if match is None:
                _fail_parse_error(ping_output)

            else:
                groups = match.groupdict()

                transmitted = int(groups['transmitted'])
                received = int(groups['received'])

                if transmitted != received:
                    _fail_lost_packets(ping_output, transmitted, received)

                else:
                    _success(ping_output)

            logger.debug(f'failed {guest_context.ping_failures} of {self.ping_threshold} allowed')

            if guest_context.ping_failures >= self.ping_threshold:
                logger.fail(f'exhausted {self.ping_threshold} ping attempts')

                self.notify(invocation, logger)

        try:
            assert invocation.guest.primary_address is not None  # narrow type

            output = tmt.utils.Command(
                'ping', '-c', str(self.ping_packets), invocation.guest.primary_address
            ).run(cwd=Path.cwd(), stream_output=False, logger=logger)

            _handle_output(output.stdout or '')

        except tmt.utils.RunError as exc:
            if exc.returncode == 1:
                _handle_output(exc.stdout or '')

            else:
                _handle_output('\n'.join(render_run_exception_streams(exc.output)))

    def do_ssh_ping(
        self,
        invocation: 'TestInvocation',
        guest_context: GuestContext,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Perform a "SSH ping" check
        """

        assert isinstance(invocation.guest, tmt.steps.provision.GuestSsh)

        logger.debug('checking SSH port', level=4)

        log = render_report_path(invocation)

        def _fail_unknown(ncat_output: str) -> None:
            """
            Handle unknown failures
            """

            logger.fail('unknown error')

            guest_context.ssh_ping_failures += 1

            report_progress(
                log,
                'ssh-ping',
                [
                    '# unknown error',
                    f'# failed {guest_context.ssh_ping_failures}'
                    f' of {self.ssh_ping_threshold} allowed',
                ],
                command_output=ncat_output,
            )

        def _fail_connection_refused(ncat_output: str) -> None:
            """
            Handle failed connection
            """

            logger.fail('connection refused')

            guest_context.ssh_ping_failures += 1

            report_progress(
                log,
                'ssh-ping',
                [
                    '# connection refused',
                    f'# failed {guest_context.ssh_ping_failures}'
                    f' of {self.ssh_ping_threshold} allowed',
                ],
                command_output=ncat_output,
            )

        def _success(ncat_output: str) -> None:
            """
            Handle successful response
            """

            logger.verbose('Received successful response to SSH ping.', level=2)

            report = ['# successful response']

            if guest_context.ssh_ping_failures != 0:
                report.append(f'# replenished failure budget back to {self.ssh_ping_threshold}')

            guest_context.ssh_ping_failures = 0

            report_progress(log, 'ssh-ping', report, command_output=ncat_output)

        try:
            assert invocation.guest.primary_address is not None  # narrow type

            output = tmt.utils.Command(
                'nc', '-zv', invocation.guest.primary_address, str(invocation.guest.port or 22)
            ).run(cwd=Path.cwd(), stream_output=False, logger=logger)

            _success(output.stderr or '')

        except tmt.utils.RunError as exc:
            if exc.returncode == 1:
                _fail_connection_refused(exc.stderr or '')

            else:
                _fail_unknown('\n'.join(render_run_exception_streams(exc.output)))

        logger.debug(
            f'failed {guest_context.ssh_ping_failures} of {self.ssh_ping_threshold} allowed'
        )

        if guest_context.ssh_ping_failures >= self.ssh_ping_threshold:
            logger.fail(f'exhausted {self.ssh_ping_threshold} SSH ping attempts')

            self.notify(invocation, logger)


@provides_check('watchdog')
class Watchdog(CheckPlugin[WatchdogCheck]):
    #
    # This plugin docstring has been reviewed and updated to follow
    # our documentation best practices. When changing it, please make
    # sure new changes are following them as well.
    #
    # https://tmt.readthedocs.io/en/stable/contribute.html#docs
    #
    """
    Take various actions when guest becomes unresponsive.

    Watchdog runs selected probes every now and then, and when a given
    number of `probes` fail, watchdog would run one or more of the
    predefined `actions`.

    Check comes with two probes, "ping" and "SSH ping", and single
    action, "reboot".

    * "ping" uses the classic ICMP echo to check whether the guest is
      still up and running,
    * "SSH ping" tries to establish SSH connection,
    * "reboot" action issues a hard reboot of the guest.

    .. warning::

        Be aware that this feature may be limited depending on how the
        guest was provisioned. See :ref:`/plugins/provision/hard-reboot`.

    Each probe has a "budget" of allowed failures, and when it runs out,
    the action is taken. A successful probe replenishes its budget to
    the original level.

    Multiple probes can be enabled at the same time, for the action to
    happen it's enough if just one of them runs out of its budget.

    .. code-block:: yaml

        check:
          - how: watchdog
            ping: true
            reboot: true

    .. code-block:: yaml

        check:
          - how: watchdog

            # Use only SSH ping.
            ping: false
            ssh-ping: true

            # Try every 5 minutes, allow 7 failed attempts, and reboot
            # the guest when we run out of attempts.
            interval: 300
            reboot: true
            ssh-ping-threshold: 7

    .. versionadded:: 1.32
    """

    _check_class = WatchdogCheck

    @classmethod
    def before_test(
        cls,
        *,
        check: WatchdogCheck,
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        # Setup a logger
        watchdog_logger = logger.clone()
        watchdog_logger.labels.append('watchdog')

        # Create a guest context for the guest we've been given
        invocation.check_data[check.how] = GuestContext()

        guest_context: GuestContext = invocation.check_data[check.how]

        if check.ping and not isinstance(invocation.guest, PINGABLE_GUEST_CLASSES):
            watchdog_logger.warning('Ping against this guest is not supported, disabling.')

            check.ping = False

        if check.ssh_ping and not isinstance(invocation.guest, SSH_PINGABLE_GUEST_CLASSES):
            watchdog_logger.warning('SSH ping against this guest is not supported, disabling.')

            check.ssh_ping = False

        def watchdog(guest_context: GuestContext) -> None:
            """
            Watchdog thread code
            """

            tid = threading.get_ident()

            watchdog_logger.debug(f'Watchdog starts in thread {tid}')

            while guest_context.keep_running:
                if check.ping:
                    check.do_ping(invocation, guest_context, watchdog_logger)

                if check.ssh_ping:
                    check.do_ssh_ping(invocation, guest_context, watchdog_logger)

                time.sleep(check.interval)

            watchdog_logger.debug(f'Watchdog finished in thread {tid}')

        guest_context.thread = threading.Thread(
            target=watchdog,
            args=(guest_context,),
            name=f'watchdog-{invocation.guest.name}',
            daemon=True,
        )

        guest_context.thread.start()

        return []

    @classmethod
    def after_test(
        cls,
        *,
        check: WatchdogCheck,
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[CheckResult]:
        watchdog_logger = logger.clone()
        watchdog_logger.labels.append('watchdog')

        guest_context: GuestContext = invocation.check_data[check.how]

        if guest_context.thread:
            watchdog_logger.debug(f'Terminating watchdog in thread {guest_context.thread.ident}')

            guest_context.keep_running = False
            guest_context.thread.join()

            guest_context.thread = None

        assert invocation.phase.step.workdir is not None  # narrow type

        return [
            CheckResult(
                name='watchdog',
                result=ResultOutcome.PASS,
                log=[render_report_path(invocation).relative_to(invocation.phase.step.workdir)],
            )
        ]



import dataclasses
import inspect
import re
from collections.abc import Iterator
from typing import Any, Callable, Optional, cast

import tmt
import tmt.base
import tmt.container
import tmt.log
import tmt.options
import tmt.steps
import tmt.steps.prepare
import tmt.steps.provision
import tmt.utils
from tmt.container import container
from tmt.plugins import PluginRegistry
from tmt.result import PhaseResult
from tmt.steps.provision import Guest
from tmt.utils import Path
from tmt.utils.templates import render_template

FEATURE_PLAYEBOOK_DIRECTORY = tmt.utils.resource_files('steps/prepare/feature')

FeatureClass = type['FeatureBase']
_FEATURE_PLUGIN_REGISTRY: PluginRegistry[FeatureClass] = PluginRegistry('prepare.feature')


#: A pattern for matching module-like keys in Ansible playbooks.
ANSIBLE_MODULE_NAME_PATTERN = re.compile(r'.+\..+\..+')

#: A template for "used modules" note for plugin docstrings.
ANSIBLE_MODULE_NOTE_TEMPLATE = """
.. note::

    This plugin requires the following Ansible modules be installed
    on the runner:

    {% for module in MODULES %}
    * `{{ module }}`__
    {% endfor %}

    {% for module in MODULES %}
        {% set module_components = module.split('.', 2) %}
    __ https://docs.ansible.com/ansible/latest/collections/{{ module_components[0] }}/{{ module_components[1] }}/{{ module_components[2] }}_module.html
    {% endfor %}
"""  # noqa: E501


def _collect_playbook_modules(feature_cls: FeatureClass, logger: tmt.log.Logger) -> set[str]:
    """
    Find all module-like keys in feature's Ansible playbooks.

    Module-like keys are keys in dictionaries that match the pattern of
    ``foo.bar.baz``, as we enforce fully-qualified module names to be
    used.
    """

    # Find module-like keys in a given object, recursively.
    def _collect_from_object(obj: Any) -> Iterator[str]:
        if isinstance(obj, list):
            for item in obj:
                yield from _collect_from_object(item)

        elif isinstance(obj, dict):
            for key, value in obj.items():
                if ANSIBLE_MODULE_NAME_PATTERN.match(key):
                    yield key

                else:
                    yield from _collect_from_object(value)

    # Find module-like keys in playbooks, inspecting one by one.
    def _collect_from_playbooks(playbooks: set[str]) -> Iterator[str]:
        for playbook_filename in playbooks:
            playbook_filepath = feature_cls._find_playbook(playbook_filename, logger)

            if playbook_filepath is None:
                continue

            yield from _collect_from_object(tmt.utils.yaml_to_list(playbook_filepath.read_text()))

    return set(_collect_from_playbooks(feature_cls.PLAYBOOKS))


def _add_modules_to_docstring(feature_cls: FeatureClass, logger: tmt.log.Logger) -> None:
    """
    Add a list of Ansible modules used by feature's playbooks to its docstring.
    """

    if not feature_cls.__doc__:
        return

    modules = _collect_playbook_modules(feature_cls, logger)

    if not modules:
        return

    feature_cls.__doc__ += '\n' + render_template(ANSIBLE_MODULE_NOTE_TEMPLATE, MODULES=modules)


def provides_feature(feature: str) -> Callable[[FeatureClass], FeatureClass]:
    """
    A decorator for registering feature plugins.

    Decorate a feature plugin class to register a feature:

    .. code-block:: python

        @provides_feature('foo')
        class Foo(ToggleableFeature):
            ...

    Decorator also inspects plugins :py:attr:`FeatureBase.PLAYBOOKS`,
    gathers all Ansible modules from listed playbooks, and adds a note
    to plugin's docstring reminding user about their necessity on the
    runner.
    """

    def _provides_feature(feature_cls: FeatureClass) -> FeatureClass:
        logger = tmt.log.Logger.get_bootstrap_logger()

        feature_cls.FEATURE_NAME = feature

        _FEATURE_PLUGIN_REGISTRY.register_plugin(
            plugin_id=feature,
            plugin=feature_cls,
            logger=logger,
        )

        if feature_cls.PLAYBOOKS:
            _add_modules_to_docstring(feature_cls, logger)

        return feature_cls

    return _provides_feature


def find_plugin(name: str) -> 'FeatureClass':
    """
    Find a plugin by its name.

    :raises GeneralError: when the plugin does not exist.
    """

    plugin = _FEATURE_PLUGIN_REGISTRY.get_plugin(name)

    if plugin is None:
        raise tmt.utils.GeneralError(
            f"Feature plugin '{name}' was not found in the feature registry."
        )

    return plugin


@container
class PrepareFeatureData(tmt.steps.prepare.PrepareStepData):
    pass


class FeatureBase(tmt.utils.Common):
    """Base class for ``feature`` plugins"""

    FEATURE_NAME: str
    PLAYBOOKS: set[str] = set()

    #: Plugin's data class listing keys this feature plugin accepts.
    #: It is eventually composed together with other feature plugins
    #: into a single class, :py:class:`PrepareFeatureData`.
    _data_class: type[PrepareFeatureData] = PrepareFeatureData

    @classmethod
    def get_data_class(cls) -> type[PrepareFeatureData]:
        """
        Return step data class for this plugin.

        By default, :py:attr:`_data_class` is returned, but plugin may
        override this method to provide different class.
        """

        return cls._data_class

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def _find_playbook(cls, filename: str, logger: tmt.log.Logger) -> Optional[Path]:
        filepath = FEATURE_PLAYEBOOK_DIRECTORY / filename
        if filepath.exists():
            return filepath

        filepath = Path(inspect.getfile(cls)).parent / filename
        if filepath.exists():
            return filepath

        logger.warning(f"Cannot find any suitable playbook for '{filename}'.", 0)
        return None

    @classmethod
    def _run_playbook(
        cls,
        op: str,
        playbook_filename: str,
        guest: Guest,
        logger: tmt.log.Logger,
    ) -> None:
        playbook_path = cls._find_playbook(playbook_filename, logger)
        if not playbook_path:
            raise tmt.utils.GeneralError(
                f"{op.capitalize()} {cls.FEATURE_NAME.upper()} is not supported on this guest."
            )

        logger.info(f'{op.capitalize()} {cls.FEATURE_NAME.upper()}')
        guest.ansible(playbook_path)


class ToggleableFeature(FeatureBase):
    """Base class for ``feature`` plugins that enable/disable a feature"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def enable(cls, guest: Guest, logger: tmt.log.Logger) -> None:
        raise NotImplementedError

    @classmethod
    def disable(cls, guest: Guest, logger: tmt.log.Logger) -> None:
        raise NotImplementedError


class Feature(FeatureBase):
    """Base class for ``feature`` plugins that enable a feature"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def enable(cls, guest: Guest, value: str, logger: tmt.log.Logger) -> None:
        raise NotImplementedError


@tmt.steps.provides_method('feature')
class PrepareFeature(tmt.steps.prepare.PreparePlugin[PrepareFeatureData]):
    """
    Easily enable and disable common features

    The ``feature`` plugin provides a comfortable way to enable
    and disable some commonly used functionality such as enabling
    and disabling the ``epel`` repository or the ``fips`` mode.

    .. note::

       The plugin requires a working Ansible to be available on the
       test runner.

    .. warning::

       The plugin may be a subject of various limitations, imposed by
       the fact it uses Ansible to implement some of the features:

       * Ansible 2.17+ no longer supports Python 3.6 and older. Guests
         where Python 3.7+ is not available cannot be prepared with the
         ``feature`` plugin. This has been observed when Fedora Rawhide
         runner is used with CentOS 7 or CentOS Stream 8 guests. Possible
         workarounds: downgrade Ansible tmt uses, or install Python 3.7+
         before using ``feature`` plugin from an alternative repository
         or local build.

    .. code-block:: yaml

        prepare:
            how: feature
            epel: disabled
            crb: enabled
            fips: enabled
            ...

    .. code-block:: shell

        prepare --how feature --epel disabled --crb enabled --fips enabled ...

    .. note::

       Features available via this plugin are implemented and shipped as
       plugins too. The list of available features and configuration keys
       will depend on which plugins you have installed.
    """

    _data_class = PrepareFeatureData

    @classmethod
    def get_data_class(cls) -> type[PrepareFeatureData]:
        """
        Return step data class for this plugin.

        ``prepare/feature`` builds the class in a dynamic way: class'
        fields are defined by discovered feature plugins. Plugins define
        their own data classes, these are collected, their fields
        extracted and merged together with the base data class fields
        (``name``, ``order``, ...) into the final data class of
        ``prepare/feature`` plugin.
        """

        # If this class' data class is not `PrepareFeatureData` anymore,
        # it means this method already constructed the dynamic class.
        if cls._data_class == PrepareFeatureData:
            # Collect fields in the base class, we must filter them out
            # from classes returned by plugins. These fields will be
            # provided by the base class, and repeating them would raise
            # an exception.
            baseclass_fields = list(tmt.container.container_fields(PrepareFeatureData))
            baseclass_field_names = [field.name for field in baseclass_fields]

            component_fields = [
                field
                for plugin in _FEATURE_PLUGIN_REGISTRY.iter_plugins()
                for field in tmt.container.container_fields(plugin.get_data_class())
                if field.name not in baseclass_field_names
            ]

            cls._data_class = cast(
                type[PrepareFeatureData],
                dataclasses.make_dataclass(
                    'PrepareFeatureData',
                    [(field.name, field.type, field) for field in component_fields],
                    bases=(PrepareFeatureData,),
                ),
            )

        return cls._data_class

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Prepare the guests
        """

        results = super().go(guest=guest, environment=environment, logger=logger)

        # Nothing to do in dry mode
        if self.opt('dry'):
            return []

        for plugin_id in _FEATURE_PLUGIN_REGISTRY.iter_plugin_ids():
            plugin_class = find_plugin(plugin_id)

            value = cast(Optional[str], getattr(self.data, plugin_class.FEATURE_NAME, None))
            if value is None:
                continue

            if issubclass(plugin_class, ToggleableFeature):
                value = value.lower()

                if value == 'enabled':
                    plugin_class.enable(guest, logger)
                elif value == 'disabled':
                    plugin_class.disable(guest, logger)
                else:
                    raise tmt.utils.GeneralError(f"Unknown plugin setting '{value}'.")

            elif issubclass(plugin_class, Feature):
                plugin_class.enable(guest, value, logger)

            else:
                raise tmt.utils.GeneralError(f"Unknown plugin implementation '{plugin_class}'.")

        return results

    def essential_requires(self) -> list[tmt.base.Dependency]:
        """
        Collect all essential requirements of the plugin.

        Essential requirements of a plugin are necessary for the plugin to
        perform its basic functionality.

        :returns: a list of requirements.
        """

        return tmt.steps.provision.essential_ansible_requires()



import os
import shutil
import sys
import tempfile

import _pytest.monkeypatch
import pytest

import tmt.cli._root
import tmt.log
from tests import CliRunner
from tmt.container import container
from tmt.utils import Path

# Prepare path to examples
PATH = Path(__file__).resolve().parent


def example(name):
    """
    Return path to given example
    """

    return PATH / "../../examples/" / name


runner = CliRunner()


def test_mini():
    """
    Minimal smoke test
    """

    tmp = tempfile.mkdtemp()
    result = runner.invoke(
        tmt.cli._root.main, ['--root', example('mini'), 'run', '-i', tmp, '-dv', 'discover']
    )
    assert result.exit_code == 0
    assert 'Found 1 plan.' in result.output
    assert '1 test selected' in result.output
    assert '/ci' in result.output
    shutil.rmtree(tmp)


def test_init():
    """
    Tree initialization
    """

    tmp = tempfile.mkdtemp()
    original_directory = os.getcwd()
    os.chdir(tmp)
    result = runner.invoke(tmt.cli._root.main, ['init'])
    assert 'Initialized the fmf tree root' in result.output
    result = runner.invoke(tmt.cli._root.main, ['init'])
    assert 'already exists' in result.output
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'mini'])
    assert 'plans/example' in result.output
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'mini'])
    assert result.exception
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'full', '--force'])
    assert 'overwritten' in result.output
    # tmt init --template mini in a clean directory
    os.system('rm -rf .fmf *')
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'mini'])
    assert 'plans/example' in result.output
    # tmt init --template full in a clean directory
    os.system('rm -rf .fmf *')
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'full'])
    assert 'tests/example' in result.output
    os.chdir(original_directory)
    shutil.rmtree(tmp)


def test_create():
    """
    Test, plan and story creation
    """

    # Create a test directory
    tmp = tempfile.mkdtemp()
    original_directory = os.getcwd()
    os.chdir(tmp)
    # Commands to test
    commands = [
        'init',
        'test create -t beakerlib test',
        'test create -t shell test',
        'plan create -t mini test',
        'plan create -t full test',
        'story create -t mini test',
        'story create -t full test',
    ]
    for command in commands:
        result = runner.invoke(tmt.cli._root.main, command.split())
        assert result.exit_code == 0
        os.system('rm -rf *')
    # Test directory cleanup
    os.chdir(original_directory)
    shutil.rmtree(tmp)


def test_step():
    """
    Select desired step
    """

    for step in ['discover', 'provision', 'prepare']:
        tmp = tempfile.mkdtemp()
        result = runner.invoke(
            tmt.cli._root.main,
            ['--feeling-safe', '--root', example('local'), 'run', '-i', tmp, step],
        )
        assert result.exit_code == 0
        assert step in result.output
        assert 'finish' not in result.output
        shutil.rmtree(tmp)


def test_step_execute():
    """
    Test execute step
    """

    tmp = tempfile.mkdtemp()
    step = 'execute'

    result = runner.invoke(
        tmt.cli._root.main, ['--root', example('local'), 'run', '-i', tmp, step]
    )

    # Test execute empty with discover output missing
    assert result.exit_code != 0
    assert isinstance(result.exception, tmt.utils.GeneralError)
    assert len(result.exception.causes) == 1
    assert isinstance(result.exception.causes[0], tmt.utils.ExecuteError)
    assert step in result.output
    assert 'provision' not in result.output
    shutil.rmtree(tmp)


def test_systemd():
    """
    Check systemd example
    """

    result = runner.invoke(tmt.cli._root.main, ['--root', example('systemd'), 'plan'])
    assert result.exit_code == 0
    assert 'Found 2 plans' in result.output
    result = runner.invoke(tmt.cli._root.main, ['--root', example('systemd'), 'plan', 'show'])
    assert result.exit_code == 0
    assert 'Tier two functional tests' in result.output


@container
class DecideColorizationTestcase:
    """
    A single test case for :py:func:`tmt.log.decide_colorization`
    """

    # Name of the testcase and expected outcome of decide_colorization()
    name: str
    expected: tuple[bool, bool]

    # Testcase environment setup to perform before calling decide_colorization()
    set_no_color_option: bool = False
    set_force_color_option: bool = False
    set_no_color_envvar: bool = False
    set_tmt_no_color_envvar: bool = False
    set_tmt_force_color_envvar: bool = False
    simulate_tty: bool = False


_DECIDE_COLORIZATION_TESTCASES = [
    # With TTY simulated
    DecideColorizationTestcase('tty, autodetection', (True, True), simulate_tty=True),
    DecideColorizationTestcase(
        'tty, disable with option', (False, False), set_no_color_option=True, simulate_tty=True
    ),
    DecideColorizationTestcase(
        'tty, disable with NO_COLOR', (False, False), set_no_color_envvar=True, simulate_tty=True
    ),
    DecideColorizationTestcase(
        'tty, disable with TMT_NO_COLOR',
        (False, False),
        set_tmt_no_color_envvar=True,
        simulate_tty=True,
    ),
    DecideColorizationTestcase(
        'tty, force with option', (True, True), set_force_color_option=True, simulate_tty=True
    ),
    DecideColorizationTestcase(
        'tty, force with TMT_FORCE_COLOR',
        (True, True),
        set_tmt_force_color_envvar=True,
        simulate_tty=True,
    ),
    DecideColorizationTestcase(
        'tty, force with TMT_FORCE_COLOR over NO_COLOR',
        (True, True),
        set_tmt_force_color_envvar=True,
        set_no_color_envvar=True,
    ),
    DecideColorizationTestcase(
        'tty, force with TMT_FORCE_COLOR over --no-color',
        (True, True),
        set_tmt_force_color_envvar=True,
        set_no_color_option=True,
    ),
    # With TTY not simulated, streams are captured
    DecideColorizationTestcase('not tty, autodetection', (False, False)),
    DecideColorizationTestcase(
        'not tty, disable with option', (False, False), set_no_color_option=True
    ),
    DecideColorizationTestcase(
        'not tty, disable with NO_COLOR', (False, False), set_no_color_envvar=True
    ),
    DecideColorizationTestcase(
        'not tty, disable with TMT_NO_COLOR', (False, False), set_tmt_no_color_envvar=True
    ),
    DecideColorizationTestcase(
        'not tty, force with option', (True, True), set_force_color_option=True
    ),
    DecideColorizationTestcase(
        'not tty, force with TMT_FORCE_COLOR', (True, True), set_tmt_force_color_envvar=True
    ),
    DecideColorizationTestcase(
        'not tty, force with TMT_FORCE_COLOR over NO_COLOR',
        (True, True),
        set_tmt_force_color_envvar=True,
        set_tmt_no_color_envvar=True,
    ),
    DecideColorizationTestcase(
        'not tty, force with TMT_FORCE_COLOR over --no-color',
        (True, True),
        set_tmt_force_color_envvar=True,
        set_no_color_option=True,
    ),
]


@pytest.mark.parametrize(
    'testcase',
    list(_DECIDE_COLORIZATION_TESTCASES),
    ids=[testcase.name for testcase in _DECIDE_COLORIZATION_TESTCASES],
)
def test_decide_colorization(
    testcase: DecideColorizationTestcase, monkeypatch: _pytest.monkeypatch.MonkeyPatch
) -> None:
    monkeypatch.delenv('NO_COLOR', raising=False)
    monkeypatch.delenv('TMT_NO_COLOR', raising=False)
    monkeypatch.delenv('TMT_FORCE_COLOR', raising=False)

    no_color = bool(testcase.set_no_color_option)
    force_color = bool(testcase.set_force_color_option)

    if testcase.set_no_color_envvar:
        monkeypatch.setenv('NO_COLOR', '')

    if testcase.set_tmt_no_color_envvar:
        monkeypatch.setenv('TMT_NO_COLOR', '')

    if testcase.set_tmt_force_color_envvar:
        monkeypatch.setenv('TMT_FORCE_COLOR', '')

    monkeypatch.setattr(sys.stdout, 'isatty', lambda: testcase.simulate_tty)
    monkeypatch.setattr(sys.stderr, 'isatty', lambda: testcase.simulate_tty)

    assert tmt.log.decide_colorization(no_color, force_color) == testcase.expected



"""
Hints for users when facing installation-related issues.

Plugins, steps, and tmt code in general can register hints to be shown
to user when an important (or optional, but interesting) package is not
available.

Hints are shown when importing plugins fails, and rendered as part of
both their CLI help and HTML documentation.

Hints are registered, each having its own ID. A hint may take advantage
of ReST to format its content for better readability.

Hints should follow the rules of docstrings, i.e. provide a short summary
on the first line, followed by more details in the rest of the text.
"""

# NOTE (happz): in my plan, this module would be an unfinished, staging
# area for hints; eventually, I would like them to be managed under the
# umbrella of `tmt about` subcommand. `print_hint()` would still exist,
# but `tmt about` would be responsible for handling hints, therefore the
# code below may change, the concept should not. And hints would cover
# wider area, e.g. describing common errors and issues, not just when
# a package is missing. They would be coupled with exceptions tmt
# raises, providing more info on command-line and in HTML docs.

import functools
import textwrap
from collections.abc import Iterator

import tmt.container
import tmt.log
import tmt.utils
import tmt.utils.rest
from tmt.log import Logger


@tmt.container.container
class Hint:
    id: str
    text: str

    @functools.cached_property
    def summary(self) -> str:
        return self.text.splitlines()[0]

    @functools.cached_property
    def ref(self) -> str:
        return f'For more details, see ``tmt about {self.id}``.'

    @functools.cached_property
    def summary_ref(self) -> str:
        # TODO: once `tmt about` will be able to print hints, we shall
        # add ref as well.
        return self.summary

    def __init__(self, hint_id: str, text: str) -> None:
        self.id = hint_id
        self.text = textwrap.dedent(text).strip()

    def _render(self, s: str, logger: Logger) -> str:
        if tmt.utils.rest.REST_RENDERING_ALLOWED:
            return tmt.utils.rest.render_rst(s, logger)

        return s

    def render_summary(self, logger: Logger) -> str:
        return self._render(self.summary, logger)

    def render_summary_ref(self, logger: Logger) -> str:
        return self._render(self.summary_ref, logger)

    def render(self, logger: Logger) -> str:
        return self._render(self.text, logger)


HINTS: dict[str, Hint] = {
    _hint_id: Hint(_hint_id, _hint)
    for _hint_id, _hint in {
        'provision': """
            You can use the ``local`` method to execute tests directly on your localhost.

            See ``tmt run provision --help`` for all available ``provision`` options.
            """,
        "report": """
            You can use the ``display`` method to show test results on the terminal.

            See ``tmt run report --help`` for all available report options.
            """,
        'ansible-not-available': """
            Make sure ``ansible-playbook`` is installed, it is required for preparing guests using
            Ansible playbooks.

            To quickly test ``ansible-playbook`` presence, you can try running
            ``ansible-playbook --help``.

            * Users who installed tmt from system repositories should install ``ansible-core``
            package.
            * Users who installed tmt from PyPI should install ``tmt[ansible]`` extra.
            """,
        # TODO: once `minute` plugin provides its own hints, we can drop
        # this hint and move it to the plugin.
        'provision/minute': """
            Make sure ``tmt-redhat-provision-minute`` package is installed, it is required for
            guests backed by 1minutetip OpenStack as provided by ``provision/minute`` plugin. The
            package is available from the internal COPR repository only.
            """,
    }.items()
}


def register_hint(id_: str, hint: str) -> None:
    """
    Register a hint for users.

    :param id_: step name for step-specific hints,
        ``<step name>/<plugin name>`` for plugin-specific hints,
        or an arbitrary string.
    :param hint: a hint to register.
    """

    if id_ in HINTS:
        raise tmt.utils.GeneralError(
            f"Registering hint '{id_}' collides with an already registered hint."
        )

    HINTS[id_] = Hint(id_, hint)


def get_hints(*ids: str, ignore_missing: bool = False) -> list[Hint]:
    """
    Find given hints.

    :param ids: ids of hints to retrieve.
    :param ignore_missing: if not set, non-existent hints will
        raise an exception. Otherwise, non-existent hints will
        be skipped.
    :returns: found hints.
    """

    if ignore_missing:

        def _get_optional_hints() -> Iterator[Hint]:
            for id_ in ids:
                hint = HINTS.get(id_)

                if hint is None:
                    continue

                yield hint

        return list(_get_optional_hints())

    def _render_mandatory_hints() -> Iterator[Hint]:
        for id_ in ids:
            hint = HINTS.get(id_)

            if hint is None:
                raise tmt.utils.GeneralError(f"Could not find hint '{id_}'.")

            yield hint

    return list(_render_mandatory_hints())


def print_hints(*ids: str, ignore_missing: bool = False, logger: tmt.log.Logger) -> None:
    """
    Display given hints by printing them as info-level messages.

    :param ids: ids of hints to render.
    :param ignore_missing: if not set, non-existent hints will
        raise an exception.
    :param logger: to use for logging.
    """

    hints = get_hints(*ids, ignore_missing=ignore_missing)

    for hint in hints:
        logger.info('hint', hint.render(logger), color='blue')



import os
import re
from typing import Any
from unittest.mock import MagicMock

import _pytest.logging
import pytest
from pytest_container.container import ContainerData

from tmt.log import Logger
from tmt.steps.provision import Guest, GuestData, GuestSsh, GuestSshData, Provision
from tmt.steps.provision.podman import GuestContainer
from tmt.utils import Command, CommandOutput, Path, RunError, ShellScript

from . import TEST_CONTAINERS


def test_multihost_name(root_logger: Logger) -> None:
    assert (
        Guest(logger=root_logger, name='foo', data=GuestData(primary_address='bar')).multihost_name
        == 'foo'
    )

    assert (
        Guest(
            logger=root_logger, name='foo', data=GuestData(primary_address='bar', role='client')
        ).multihost_name
        == 'foo (client)'
    )


@pytest.mark.parametrize(
    ('stdout', 'expected'),
    [
        (
            # no-connection-closed
            os.linesep + 'last-line' + os.linesep,
            os.linesep + 'last-line' + os.linesep,
        ),
        (
            # connection-closed-not-last-line
            os.linesep + 'Connection to 127.0.0.1 closed.' + os.linesep + 'last-line' + os.linesep,
            os.linesep + 'Connection to 127.0.0.1 closed.' + os.linesep + 'last-line' + os.linesep,
        ),
        (
            # connection-closed
            os.linesep + 'some-line' + os.linesep + 'Connection to 127.0.0.1 closed.' + os.linesep,
            os.linesep + 'some-line' + os.linesep,
        ),
        (
            # shared-connection-closed
            os.linesep
            + 'some-line'
            + os.linesep
            + 'Shared connection to 127.0.0.1 closed.'
            + os.linesep,
            os.linesep + 'some-line' + os.linesep,
        ),
    ],
    ids=(
        'no-connection-closed',
        'connection-closed-not-last-line',
        'connection-closed',
        'shared-connection-closed',
    ),
)
def test_execute_no_connection_closed(
    root_logger: Logger, stdout: str, expected: str, monkeypatch: Any
) -> None:
    step = Provision(plan=MagicMock(name='mock<plan>'), data={}, logger=root_logger)
    guest = GuestSsh(
        logger=root_logger, parent=step, name='foo', data=GuestSshData(primary_address='bar')
    )

    monkeypatch.setattr(
        guest,
        '_run_guest_command',
        MagicMock(return_value=CommandOutput(stdout=stdout, stderr=None)),
    )

    output = guest.execute(Command('some-command'), test_session=True)
    assert output.stdout == expected

    output = guest.execute(Command('some-command'))
    assert output.stdout == stdout


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container',),  # noqa: PT006
    [(TEST_CONTAINERS[container_name],) for container_name in sorted(TEST_CONTAINERS.keys())],
    indirect=["container"],
    ids=[TEST_CONTAINERS[container_name].url for container_name in sorted(TEST_CONTAINERS.keys())],
)
def test_mkdtemp(
    container: ContainerData,
    guest: GuestContainer,
    root_logger: Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    guest.execute(ShellScript('mkdir -p /tmp/qux'))

    with guest.mkdtemp(
        prefix='bar',
        template='XXXXXXbazXXXXXX',
        parent=Path('/tmp/qux'),
    ) as path:
        guest.execute(ShellScript(f'ls -al {path}'))

    with pytest.raises(RunError) as exc_context:
        guest.execute(ShellScript(f'ls -al {path}'))

    assert re.match(r'(?i).*?No such file or directory.*', exc_context.value.stderr or '')



from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

import click
from fmf.utils import listed

import tmt
from tmt.container import container, field, key_to_option

if TYPE_CHECKING:
    import tmt.cli
    import tmt.options
    import tmt.steps

import tmt.base
import tmt.steps
import tmt.utils
from tmt.options import option
from tmt.plugins import PluginRegistry
from tmt.steps import Action
from tmt.utils import GeneralError, Path


@container
class TestOrigin:
    """
    Describes the origin of a test.
    """

    #: Name of the ``discover`` phase that added the test.
    phase: str

    #: The test in question.
    test: 'tmt.Test'


@container
class DiscoverStepData(tmt.steps.WhereableStepData, tmt.steps.StepData):
    dist_git_source: bool = field(
        default=False,
        option='--dist-git-source',
        is_flag=True,
        help='Download DistGit sources and ``rpmbuild -bp`` them (can be skipped).',
    )

    # TODO: use enum!
    dist_git_type: Optional[str] = field(
        default=None,
        option='--dist-git-type',
        choices=tmt.utils.git.get_distgit_handler_names,
        help="""
            Use the provided DistGit handler instead of the auto detection.
            Useful when running from forked repositories.
            """,
    )

    dist_git_download_only: bool = field(
        default=False,
        option="--dist-git-download-only",
        is_flag=True,
        help="Just download the sources. No ``rpmbuild -bp``, "
        "nor installation of require or buildddeps happens.",
    )

    dist_git_install_builddeps: bool = field(
        default=False,
        option="--dist-git-install-builddeps",
        is_flag=True,
        help="Install package build dependencies according to the specfile.",
    )

    dist_git_require: list['tmt.base.DependencySimple'] = field(
        default_factory=list,
        option="--dist-git-require",
        metavar='PACKAGE',
        multiple=True,
        help="""
            Additional required package to be present before sources are prepared.
            The ``rpm-build`` package itself is installed automatically.
            """,
        # *simple* requirements only
        normalize=lambda key_address, value, logger: tmt.base.assert_simple_dependencies(
            tmt.base.normalize_require(key_address, value, logger),
            "'dist_git_require' can be simple packages only",
            logger,
        ),
        serialize=lambda packages: [package.to_spec() for package in packages],
        unserialize=lambda serialized: [
            tmt.base.DependencySimple.from_spec(package) for package in serialized
        ],
    )


DiscoverStepDataT = TypeVar('DiscoverStepDataT', bound=DiscoverStepData)


class DiscoverPlugin(tmt.steps.GuestlessPlugin[DiscoverStepDataT, None]):
    """
    Common parent of discover plugins
    """

    # ignore[assignment]: as a base class, DiscoverStepData is not included in
    # DiscoverStepDataT.
    _data_class = DiscoverStepData  # type: ignore[assignment]

    # Methods ("how: ..." implementations) registered for the same step.
    _supported_methods: PluginRegistry[tmt.steps.Method] = PluginRegistry('step.discover')

    @classmethod
    def base_command(
        cls,
        usage: str,
        method_class: Optional[type[click.Command]] = None,
    ) -> click.Command:
        """
        Create base click command (common for all discover plugins)
        """

        # Prepare general usage message for the step
        if method_class:
            usage = Discover.usage(method_overview=usage)

        # Create the command
        @click.command(cls=method_class, help=usage)
        @click.pass_context
        @option('-h', '--how', metavar='METHOD', help='Use specified method to discover tests.')
        @tmt.steps.PHASE_OPTIONS
        def discover(context: 'tmt.cli.Context', **kwargs: Any) -> None:
            context.obj.steps.add('discover')
            Discover.store_cli_invocation(context)

        return discover

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Perform actions shared among plugins when beginning their tasks
        """

        self.go_prolog(logger or self._logger)

    def tests(
        self, *, phase_name: Optional[str] = None, enabled: Optional[bool] = None
    ) -> list['TestOrigin']:
        """
        Return discovered tests.

        :param phase_name: if set, return only tests discovered by the
            phase of this name. Otherwise, all tests discovered by the
            phase are returned.

            .. note::

               This parameter exists to present unified interface with
               :py:meth:`tmt.steps.discover.Discover.tests` API, but it
               has no interesting effect in case of individual phases:

               * left unset, all tests discovered by the phase are
                 returned,
               * set to a phase name, tests discovered by that phase
                 should be returned. But a phase does not have access to
                 other phases' tests, therefore setting it to anything
                 but this phase name would produce an empty list.
        :param enabled: if set, return only tests that are enabled
            (``enabled=True``) or disabled (``enabled=False``). Otherwise,
            all tests are returned.
        :returns: a list of phase name and test pairs.
        """

        raise NotImplementedError

    def download_distgit_source(
        self,
        distgit_dir: Path,
        target_dir: Path,
        handler_name: Optional[str] = None,
    ) -> None:
        """
        Download sources to the target_dir

        distgit_dir is path to the DistGit repository
        """

        tmt.utils.git.distgit_download(
            distgit_dir=distgit_dir,
            target_dir=target_dir,
            handler_name=handler_name,
            caller=self,
            logger=self._logger,
        )

    def log_import_plan_details(self) -> None:
        """
        Log details about the imported plan
        """

        parent = cast(Optional[Discover], self.parent)
        if (
            parent
            and parent.plan._original_plan
            and parent.plan._original_plan._imported_plan_fmf_ids
        ):
            for remote_plan_id in parent.plan._original_plan._imported_plan_fmf_ids:
                # FIXME: cast() - https://github.com/python/mypy/issues/7981
                # Note the missing Optional for values - to_minimal_dict() would
                # not include unset keys, therefore all values should be valid.
                for key, value in cast(dict[str, str], remote_plan_id.to_minimal_spec()).items():
                    self.verbose(f'import {key}', value, 'green')

    def post_dist_git(self, created_content: list[Path]) -> None:
        """
        Discover tests after dist-git applied patches
        """


class Discover(tmt.steps.Step):
    """
    Gather information about test cases to be executed.
    """

    _plugin_base_class = DiscoverPlugin

    def __init__(
        self,
        *,
        plan: 'tmt.base.Plan',
        data: tmt.steps.RawStepDataArgument,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Store supported attributes, check for sanity
        """

        super().__init__(plan=plan, data=data, logger=logger)

        # Collection of discovered tests
        self._tests: dict[str, list[tmt.Test]] = {}
        self._failed_tests: dict[str, list[tmt.Test]] = {}

        # Test will be (re)discovered in other phases/steps
        self.extract_tests_later: bool = False

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        return {*super()._preserved_workdir_members, 'tests.yaml'}

    def load(self) -> None:
        """
        Load step data from the workdir
        """

        if self.should_run_again:
            self.debug('Run discover again when reexecuting to capture changes in plans')
        else:
            super().load()

        try:
            raw_test_data = tmt.utils.yaml_to_list(self.read(Path('tests.yaml')))

            self._tests = {}

            for raw_test_datum in raw_test_data:
                # The name of `discover` phases providing the test was added in 1.24.
                # Unfortunately, the field is required for correct work of `execute`,
                # now when it is parallel in nature. Without it, it's not possible
                # to pick the right `discover` phase which then provides the list
                # of tests to execute. Therefore raising an error instead of guessing
                # what the phase could be.
                if key_to_option('discover_phase') not in raw_test_datum:
                    # TODO: there should be a method for creating workdir-aware paths...
                    path = (
                        self.workdir / Path('tests.yaml') if self.workdir else Path('tests.yaml')
                    )

                    raise tmt.utils.BackwardIncompatibleDataError(
                        f"Could not load '{path}' whose format is not compatible "
                        "with tmt 1.24 and newer."
                    )

                phase_name = raw_test_datum.pop(key_to_option('discover_phase'))

                if phase_name not in self._tests:
                    self._tests[phase_name] = []

                self._tests[phase_name].append(
                    tmt.Test.from_dict(
                        logger=self._logger,
                        mapping=raw_test_datum,
                        name=raw_test_datum['name'],
                        skip_validation=True,
                    )
                )

        except tmt.utils.FileError:
            self.debug('Discovered tests not found.', level=2)

    def save(self) -> None:
        """
        Save step data to the workdir
        """

        super().save()

        # Create tests.yaml with the full test data
        raw_test_data: list[tmt.export._RawExportedInstance] = []

        for phase_name, phase_tests in self._tests.items():
            for test in phase_tests:
                if test.enabled is not True:
                    continue

                exported_test = test._export(include_internal=True)
                exported_test[key_to_option('discover_phase')] = phase_name

                raw_test_data.append(exported_test)

        self.write(Path('tests.yaml'), tmt.utils.dict_to_yaml(raw_test_data))

    def _discover_from_execute(self) -> None:
        """
        Check the execute step for possible shell script tests
        """

        # Check scripts for command line and data, convert to list if needed
        scripts = self.plan.execute.opt('script')
        if not scripts:
            scripts = getattr(self.plan.execute.data[0], 'script', [])
        if not scripts:
            return
        if isinstance(scripts, str):
            scripts = [scripts]

        # Avoid circular imports
        from tmt.steps.discover.shell import DiscoverShellData, TestDescription

        # Give a warning when discover step defined as well
        if self.data and not all(datum.is_bare for datum in self.data):
            raise tmt.utils.DiscoverError(
                "Use either 'discover' or 'execute' step to define tests, but not both."
            )

        if not isinstance(self.data[0], DiscoverShellData):
            # TODO: or should we rather create a new `shell` discovery step data,
            # and fill it with our tests? Before step data patch, `tests` attribute
            # was simply created as a list, with no check whether the step data and
            # plugin even support `data.tests`. Which e.g. `internal` does not.
            # Or should we find the first DiscoverShellData instance, use it, and
            # create a new one when no such entry exists yet?
            raise GeneralError(
                f'Cannot append tests from execute to non-shell step "{self.data[0].how}"'
            )

        discover_step_data = self.data[0]

        # Check the execute step for possible custom duration limit
        # FIXME: cast() - https://github.com/teemtee/tmt/issues/1540
        duration = cast(
            str, getattr(self.plan.execute.data[0], 'duration', tmt.base.DEFAULT_TEST_DURATION_L2)
        )

        # Prepare the list of tests
        for index, script in enumerate(scripts):
            name = f'script-{str(index).zfill(2)}'
            discover_step_data.tests.append(
                TestDescription(name=name, test=script, duration=duration)
            )

    def wake(self) -> None:
        """
        Wake up the step (process workdir and command line)
        """

        super().wake()

        # Check execute step for possible tests (unless already done)
        if self.status() is None:
            self._discover_from_execute()

        # Choose the right plugin and wake it up
        for data in self.data:
            # FIXME: cast() - see https://github.com/teemtee/tmt/issues/1599
            plugin = cast(
                DiscoverPlugin[DiscoverStepData], DiscoverPlugin.delegate(self, data=data)
            )
            self._phases.append(plugin)
            plugin.wake()

        # Nothing more to do if already done and not asked to run again
        if self.status() == 'done' and not self.should_run_again:
            self.debug('Discover wake up complete (already done before).', level=2)
        # Save status and step data (now we know what to do)
        else:
            self.status('todo')
            self.save()

    def summary(self) -> None:
        """
        Give a concise summary of the discovery
        """

        # Summary of selected tests
        text = listed(len(self.tests(enabled=True)), 'test') + ' selected'
        self.info('summary', text, 'green', shift=1)
        # Test list in verbose mode
        for test_origin in self.tests(enabled=True):
            self.verbose(test_origin.test.name, color='red', shift=2)

    def go(self, force: bool = False) -> None:
        """
        Discover all tests
        """

        super().go(force=force)

        # Nothing more to do if already done
        if self.status() == 'done':
            self.info('status', 'done', 'green', shift=1)
            self.summary()
            self.actions()
            return

        # Perform test discovery, gather discovered tests
        for phase in self.phases(classes=(Action, DiscoverPlugin)):
            if isinstance(phase, Action):
                phase.go()

            elif isinstance(phase, DiscoverPlugin):
                if not phase.enabled_by_when:
                    continue

                # Go and discover tests
                phase.go()

                self._tests[phase.name] = []

                # Prefix test name only if multiple plugins configured
                prefix = f'/{phase.name}' if len(self.phases()) > 1 else ''
                # Check discovered tests, modify test name/path
                for test_origin in phase.tests(enabled=True):
                    test = test_origin.test

                    test.name = f"{prefix}{test.name}"
                    test.path = Path(f"/{phase.safe_name}{test.path}")
                    # Update test environment with plan environment
                    test.environment.update(self.plan.environment)
                    self._tests[phase.name].append(test)

            else:
                raise GeneralError(f'Unexpected phase in discover step: {phase}')

        for test_origin in self.tests():
            test_origin.test.serial_number = self.plan.draw_test_serial_number(test_origin.test)

        # Show fmf identifiers for tests discovered in plan
        # TODO: This part should go into the 'fmf.py' module
        if self.opt('fmf_id'):
            if self.tests(enabled=True):
                export_fmf_ids: list[str] = []

                for test_origin in self.tests(enabled=True):
                    fmf_id = test_origin.test.fmf_id

                    if not fmf_id.url:
                        continue

                    exported = test_origin.test.fmf_id.to_minimal_spec()

                    if fmf_id.default_branch and fmf_id.ref == fmf_id.default_branch:
                        exported.pop('ref')

                    export_fmf_ids.append(tmt.utils.dict_to_yaml(exported, start=True))

                click.echo(''.join(export_fmf_ids), nl=False)
            return

        if self.should_run_again and tmt.base.Test._opt('failed_only'):
            failed_results: list[tmt.base.Result] = []
            assert self.parent is not None  # narrow type
            assert isinstance(self.parent, tmt.base.Plan)  # narrow type

            # Get failed results from previous run execute
            for result in self.parent.execute._results:
                if (
                    result.result is not tmt.result.ResultOutcome.PASS
                    and result.result is not tmt.result.ResultOutcome.INFO
                ):
                    failed_results.append(result)

            # Filter existing tests into another variable which is then used by tests() method
            for test_phase in self._tests:
                self._failed_tests[test_phase] = []
                for test in self._tests[test_phase]:
                    for result in failed_results:
                        if test.name == result.name and test.serial_number == result.serial_number:
                            self._failed_tests[test_phase].append(test)

        # Give a summary, update status and save
        self.summary()
        self.status('done')
        self.save()

    def tests(
        self, *, phase_name: Optional[str] = None, enabled: Optional[bool] = None
    ) -> list['TestOrigin']:
        """
        Return discovered tests.

        :param phase_name: if set, return only tests discovered by the
            phase of this name. Otherwise, tests discovered by all
            phases are returned.
        :param enabled: if set, return only tests that are enabled
            (``enabled=True``) or disabled (``enabled=False``). Otherwise,
            all tests are returned.
        :returns: a list of phase name and test pairs.
        """

        from tmt.steps.discover import TestOrigin

        suitable_tests = self._failed_tests if self._failed_tests else self._tests
        suitable_phases = [phase_name] if phase_name is not None else list(self._tests.keys())

        def _iter_tests() -> Iterator['TestOrigin']:
            # PLR1704: this redefinition of `phase_name` is acceptable, the original
            # value is not longer needed as it has been turned into `suitable_phases`.
            for phase_name, phase_tests in suitable_tests.items():  # noqa: PLR1704
                if phase_name not in suitable_phases:
                    continue

                for test in phase_tests:
                    yield TestOrigin(test=test, phase=phase_name)

        if enabled is None:
            return list(_iter_tests())

        return [
            test_origin for test_origin in _iter_tests() if test_origin.test.enabled is enabled
        ]



"""
Metadata linting.

Internal APIs, classes, shared functionality and helpers for test, plan and
story metadata linting.

A mixin class, :py:class:`Lintable`, provides the required functionality for
base classes. Namely, it takes care of linter discovery and provides
:py:meth:`Lintable.lint` method to run them.

Classes spiced with ``Lintable`` define their sets of linters. Consider the
following examples:

.. code-block:: python

   # Methods whose names start with ``lint_*`` prefix are considered *linters*,
   # and linters perform one or more *checks* users can enable or disable.
   def lint_path_exists(self) -> LinterReturn:
       # A linter must have a docstring which is then used to generate documentation,
       # e.g. when ``lint --list-checks`` is called. The docstring must begin with
       # a linter *id*. The id should match ``[CTPSG]\\d\\d\\d`` regular expression:
       # ``C``ommon, ``T``est, ``P``lan, ``S``tory, ``G``roup, plus a three-digit
       # serial number of the check among its peers.
       ''' T004: test directory path must exist '''

       # Linter implements a generator (see :py:member:`LinterReturn`) yielding
       # two item tuples of :py:class:`LinterOutcome` and string messages.

       if not self.path:
           yield LinterOutcome.FAIL, 'directory path is not set'
           return

       test_path = os.path.join(self.node.root, os.path.relpath(self.path.strip(), '/'))

       if not os.path.exists(test_path):
           yield LinterOutcome.FAIL, f'test path "{test_path}" does not exist'
           return

       yield LinterOutcome.PASS, f'test path "{test_path}" does exist'

   def lint_manual_valid_markdown(self) -> LinterReturn:
       ''' T008: manual test should be valid markdown '''

       # Linter should yield `SKIP` outcome when it does not apply, to announce
       # it did inspect the object but realized the object is out of scope of
       # the linter, and checks do not apply.
       if not self.manual:
            yield LinterOutcome.SKIP, 'not a manual test'
            return

       manual_test = os.path.join(self.node.root, self.path.strip())

       warnings = tmt.export.check_md_file_respects_spec(manual_test)

       if warnings:
           # Linter may yield as many tuples as it deems necessary. This allows
           # for linters iterating over more granular aspects of metadata,
           # providing more specific hints.
           for warning in warnings:
              yield LinterOutcome.WARN, warning

       ...
"""

import enum
import re
import textwrap
from collections.abc import Iterable, Iterator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    TypeVar,
)

import tmt
import tmt.utils
from tmt.container import container
from tmt.utils.themes import style

if TYPE_CHECKING:
    import tmt.base


# ignore[type-arg]: bound type vars cannot be generic, and it would create a loop anyway.
LintableT = TypeVar('LintableT', bound='Lintable')  # type: ignore[type-arg]


class LinterOutcome(enum.Enum):
    SKIP = 'skip'
    PASS = 'pass'
    FAIL = 'fail'
    WARN = 'warn'
    FIXED = 'fixed'


# TODO: these would be cool, but first we should get rid of Result
# adding color when it shouldn't have.
#
# _OUTCOME_TO_MARK = {
#     LinterOutcome.SKIP: '-',
#     LinterOutcome.PASS: '\N{heavy check mark}',
#     LinterOutcome.FAIL: '\N{heavy multiplication x}',
#     LinterOutcome.WARN: '\N{exclamation mark}',
#     LinterOutcome.FIXED: '\N{heavy check mark}'
# }

_OUTCOME_TO_MARK = {
    LinterOutcome.SKIP: 'skip',
    LinterOutcome.PASS: 'pass',
    LinterOutcome.FAIL: 'fail',
    LinterOutcome.WARN: 'warn',
    LinterOutcome.FIXED: 'fix ',
}

_OUTCOME_TO_COLOR = {
    LinterOutcome.SKIP: 'blue',
    LinterOutcome.PASS: 'green',
    LinterOutcome.FAIL: 'red',
    LinterOutcome.WARN: 'yellow',
    LinterOutcome.FIXED: 'green',
}


#: Info on how a linter decided: linter itself, its outcome & the message.
LinterRuling = tuple['Linter', LinterOutcome, LinterOutcome, str]
#: A return value type of a single linter.
LinterReturn = Iterator[tuple[LinterOutcome, str]]
#: A linter itself, a callable method.
# TODO: ignore[type-arg]: `Lintable` is a generic type, and mypy starts
# reporting it since 1.7.1 or so. Adding the parameter here would require
# a bigger patch than a mere bump of mypy version. Leaving for later.
LinterCallback = Callable[['Lintable'], LinterReturn]  # type: ignore[type-arg]

_LINTER_DESCRIPTION_PATTERN = re.compile(
    r"""
    ^                      # must match the whole string
    (?P<id>[CTPSG]\d\d\d): # check ID, the class initials & a three-digit number
    \s*                    # optional white space
    (?P<short>.+)          # check description
    $                      # must match the whole string
    """,
    re.VERBOSE,
)


@container(init=False)
class Linter:
    """
    A single linter
    """

    callback: LinterCallback
    id: str

    help: str
    description: Optional[str] = None

    def __init__(self, callback: LinterCallback) -> None:
        self.callback = callback

        if not callback.__doc__:
            raise tmt.utils.GeneralError(f"Linter '{callback}' lacks docstring.")

        match = _LINTER_DESCRIPTION_PATTERN.match(textwrap.dedent(callback.__doc__).strip())

        if not match:
            raise tmt.utils.GeneralError(f"Linter '{callback}' docstring has wrong format.")

        components = match.groupdict()

        self.id = components['id'].strip()
        self.help = components['short'].strip()

    def format(self) -> list[str]:
        """
        Format the linter for printing or logging.

        :returns: a string description of the linter, suitable for
            logging or help texts, in the form of lines of text.
        """

        return [f'{self.id}: {self.help}']


class Lintable(Generic[LintableT]):
    """
    Mixin class adding support for linting of class instances
    """

    # Declare linter registry as a class variable, but do not initialize it. If initialized
    # here, the mapping would be shared by all classes, which is not a desirable attribute.
    # Instead, mapping will be created by `get_linter_registry()`.
    _linter_registry: ClassVar[list[Linter]]

    # Keep this method around, to correctly support Python's method resolution order.
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    # Cannot use @property as this must remain classmethod
    @classmethod
    def get_linter_registry(cls) -> list[Linter]:
        """
        Return - or initialize - linter registry
        """

        if not hasattr(cls, '_linter_registry'):
            cls._linter_registry = []

        return cls._linter_registry

    @classmethod
    def discover_linters(cls) -> None:
        """
        Discover and register all linters implemented by this class.

        A linter is a method whose name starts with ``lint_`` prefix. It must
        have a docstring which serves as a hint for ``--list-checks`` output.
        """

        for name in dir(cls):
            if not name.startswith('lint_'):
                continue

            cls.get_linter_registry().append(tmt.lint.Linter(getattr(cls, name)))

    @classmethod
    def resolve_enabled_linters(
        cls,
        enable_checks: Optional[list[str]] = None,
        disable_checks: Optional[list[str]] = None,
    ) -> list[Linter]:
        """
        Produce a list of enabled linters from all registered ones.

        Method combines three inputs:

        * registered linters, acquired from the class registry,
        * list of checks to enable, and
        * list of checks to disable

        into a single list of linters that are considered as enabled.

        :param enable_checks: if set, only linters providing the listed checks
            would be included in the output.
        :param disable_checks: if set, linters providing the listed checks would
            be removed from the output.
        :returns: list of linters that were registered, and whose checks were
            enabled and not disabled.
        """

        linters: list[Linter] = []

        if not enable_checks:
            linters = cls.get_linter_registry()

        else:
            linters = []

            for needle in enable_checks:
                linters += [linter for linter in cls.get_linter_registry() if needle in linter.id]

        if disable_checks:
            linters = [linter for linter in linters if linter.id not in disable_checks]

        return linters

    def lint(
        self,
        enable_checks: Optional[list[str]] = None,
        disable_checks: Optional[list[str]] = None,
        enforce_checks: Optional[list[str]] = None,
        linters: Optional[list[Linter]] = None,
    ) -> tuple[bool, list[LinterRuling]]:
        """
        Check the instance against a battery of linters and report results.

        :param enable_checks: if set, only linters providing the listed checks
            would be applied.
        :param disable_checks: if set, linters providing the listed checks would
            not be applied.
        :param enforce_checks: if set, listed checks would be marked as failed
            if their outcome is not ``pass``, i.e. even a warning would become
            a fail.
        :param linters: if set, only these linters would be applied. Providing
            ``linters`` makes ``enable_checks`` and ``disable_checks`` ignored.
        :returns: a tuple of two items: a boolean reporting whether the instance
            passed the test, and a list of :py:class:`LinterRuling` items, each
            describing one linter outcome. Note that linters may produce none or
            more than one outcome.
        """

        enforce_checks = enforce_checks or []

        linters = (
            linters
            if linters is not None
            else self.resolve_enabled_linters(
                enable_checks=enable_checks, disable_checks=disable_checks
            )
        )

        valid = True
        rulings: list[LinterRuling] = []

        for linter in sorted(linters, key=lambda x: x.id):
            for outcome, message in linter.callback(self):
                if outcome == LinterOutcome.FAIL:
                    rulings.append((linter, outcome, outcome, message))

                    valid = False

                elif outcome != LinterOutcome.PASS:
                    if linter.id in enforce_checks:
                        rulings.append((linter, outcome, LinterOutcome.FAIL, message))

                        valid = False

                    else:
                        rulings.append((linter, outcome, outcome, message))

                else:
                    rulings.append((linter, outcome, outcome, message))

        return valid, rulings

    @classmethod
    def format_linters(cls) -> str:
        """
        Format registered linters for printing or logging.

        :returns: a string description of registered linters, suitable for
            logging or help texts.
        """

        hints: list[str] = []

        for linter in sorted(cls.get_linter_registry(), key=lambda x: x.id):
            hints += linter.format()

        return '\n'.join(hints)


def filter_allowed_checks(
    rulings: Iterable[LinterRuling],
    outcomes: Optional[list[LinterOutcome]] = None,
) -> Iterator[LinterRuling]:
    """
    Filter only rulings whose outcomes are allowed.

    :param rulings: rulings to process.
    :param outcomes: a list of allowed ruling outcomes. If not set, all outcomes
        are allowed.
    :yields: rulings with allowed outcomes.
    """

    outcomes = outcomes or []

    for linter, actual_outcome, eventual_outcome, message in rulings:
        if outcomes and eventual_outcome not in outcomes:
            continue

        yield (linter, actual_outcome, eventual_outcome, message)


def format_rulings(rulings: Iterable[LinterRuling]) -> Iterator[str]:
    """
    Format rulings for printing or logging.

    :param rulings: rulings to format.
    :yields: rulings formatted as separate strings.
    """

    # Find out whether there is a ruling whose actual outcome is not the same
    # as its eventual outcome. That means the actual outcome has been overruled,
    # waived or turned into a failure - if there is any such ruling, we should
    # display the actual => eventual transition. If there's no such ruling, we
    # can display just one outcome, they are both same, and indentation and
    # padding would be simpler.
    display_eventual = any(
        actual_outcome != eventual_outcome for _, actual_outcome, eventual_outcome, _ in rulings
    )

    for linter, actual_outcome, eventual_outcome, message in rulings:
        assert actual_outcome in _OUTCOME_TO_MARK
        assert actual_outcome in _OUTCOME_TO_COLOR

        actual_status = style(
            _OUTCOME_TO_MARK[actual_outcome], fg=_OUTCOME_TO_COLOR[actual_outcome]
        )

        if display_eventual:
            assert eventual_outcome in _OUTCOME_TO_MARK
            assert eventual_outcome in _OUTCOME_TO_COLOR

            eventual_status = style(
                _OUTCOME_TO_MARK[eventual_outcome], fg=_OUTCOME_TO_COLOR[eventual_outcome]
            )

            eventual = ' ' * 8 if actual_outcome == eventual_outcome else f' -> {eventual_status}'

        else:
            eventual = ''

        yield f'{actual_status}{eventual} {linter.id} {message}'



from tmt.container import MetadataContainer


class MrackTranslation(MetadataContainer):
    """
    Here's a full config example:

    .. code-block:: yaml

     beaker:
       translations:
         - requirement: cpu.processors
           template: '{"cpu": {"processors": {"_op": "{{ OPERATOR }}", "_value": "{{ VALUE }}"}}}'

    """

    requirement: str
    template: str


class MrackHardware(MetadataContainer):
    translations: list[MrackTranslation]


class HardwareConfig(MetadataContainer):
    beaker: MrackHardware



"""
Handle Plugins
"""

import importlib
import os
import pkgutil
import sys
from collections.abc import Iterator
from importlib.metadata import entry_points
from types import ModuleType
from typing import Any, Generic, Optional, TypeVar, cast

import tmt
import tmt.utils
from tmt.log import Logger
from tmt.utils import Path

ModuleT = TypeVar('ModuleT', bound=ModuleType)


# Two possibilities to load additional plugins:
# entry_points (setup_tools)
ENTRY_POINT_NAME = 'tmt.plugin'
# Directories with module in environment variable
ENVIRONMENT_NAME = 'TMT_PLUGINS'

# Make a note when plugins have been already explored
ALREADY_EXPLORED = False

_TMT_ROOT = Path(tmt.__file__).resolve().parent


def discover(path: Path) -> Iterator[str]:
    """
    Discover available plugins for given paths
    """

    for _, name, package in pkgutil.iter_modules([str(path)]):
        if not package:
            yield name


#
# Explore available sources, and load all plugins found. Possible sources are:
#
# * tmt's own packages (_explore_packages)
#    - tmt.steps.*
#    - tmt.export
#    - tmt.plugins
#
# * directories (_explore_directories)
#    - directories listed in TMT_PLUGINS envvar
#
# * packaging entry points (_explore_entry_points)
#    - tmt.plugin
#


# A list of tmt (sub)packages that may contain plugins. For each package, we
# track the package name and its path relative to tmt package sources.
#
# If you think of adding new package with plugins tmt should load on start,
# this is the place.
#
# TODO: this should be a list! Unfortunately, there is a circular import in play:
# tmt.steps import tmt.export, and tmt.export import tmt.steps.STEPS. Until that
# is resolved, to use `Exportable` in tmt.steps, we need a delayed import. Hence
# the function.
def _discover_packages() -> list[tuple[str, Path]]:
    from tmt.steps import STEPS

    return [(f'tmt.steps.{step}', Path('steps') / step) for step in STEPS] + [
        ('tmt.plugins', Path('plugins')),
        ('tmt.export', Path('export')),
        ('tmt.frameworks', Path('frameworks')),
        ('tmt.checks', Path('checks')),
        ('tmt.package_managers', Path('package_managers')),
        ('tmt.steps.prepare.feature', Path('steps/prepare/feature')),
        ('tmt.plugins.plan_shapers', Path('plugins/plan_shapers')),
    ]


def _explore_package(package: str, path: Path, logger: Logger) -> None:
    """
    Import plugins from a given Python package
    """

    logger.debug(f"Import plugins from the '{package}' package.")
    logger = logger.descend()

    for module in discover(path):
        import_module(module=f'{package}.{module}', logger=logger)


def _explore_directory(path: Path, logger: Logger) -> None:
    """
    Import plugins dropped into a directory
    """

    logger.debug(f"Import plugins from the '{path}' directory.")
    logger = logger.descend()

    _path = str(path)

    for module in discover(path):
        if _path not in sys.path:
            sys.path.insert(0, _path)

        import_module(module=module, path=path, logger=logger)


def _explore_custom_directories(logger: Logger) -> None:
    """
    Import plugins from directories listed in ``TMT_PLUGINS`` envvar
    """

    logger.debug('Import plugins from custom directories.')
    logger = logger.descend()

    if not os.environ.get(ENVIRONMENT_NAME):
        logger.debug(
            f"No custom directories found in the '{ENVIRONMENT_NAME}' environment variable."
        )
        return

    for _path in os.environ[ENVIRONMENT_NAME].split(os.pathsep):
        # TID251: `pathlib` does not provide `os.patch.expandvars`, its
        # use is allowed. For the simplicity, keeping `expanduser` as well,
        # to avoid str -> Path -> str -> Path conversion.
        _explore_directory(
            Path(os.path.expandvars(os.path.expanduser(_path))).resolve(),  # noqa: TID251
            logger,
        )


def _explore_entry_point(entry_point: str, logger: Logger) -> None:
    """
    Import all plugins hooked to an entry points
    """

    logger.debug(f"Import plugins from the '{entry_point}' entry point.")
    logger = logger.descend()

    try:
        eps = entry_points()
        if hasattr(eps, "select"):
            entry_point_group = eps.select(group=entry_point)
        else:
            entry_point_group = eps[entry_point]

        for found in entry_point_group:
            logger.debug(f"Loading plugin '{found.name}' ({found.value}).")
            found.load()

    except KeyError:
        logger.debug(f"No plugins detected for the '{entry_point}' entry point.")


def _explore_packages(logger: Logger) -> None:
    """
    Import all plugins bundled into tmt package
    """

    logger.debug('Import plugins from tmt packages.')

    for name, path in _discover_packages():
        _explore_package(name, _TMT_ROOT / path, logger.descend())


def _explore_directories(logger: Logger) -> None:
    """
    Import all plugins from various directories
    """

    logger.debug('Import plugins from custom directories.')

    _explore_custom_directories(logger.descend())


def _explore_entry_points(logger: Logger) -> None:
    """
    Import all plugins hooked to entry points
    """

    logger.debug('Import plugins from entry points.')

    _explore_entry_point(ENTRY_POINT_NAME, logger.descend())


def explore(logger: Logger, again: bool = False) -> None:
    """
    Explore all available plugin locations

    By default plugins are explored only once to save time. Repeated
    call does not have any effect. Use ``again=True`` to force plugin
    exploration even if it has been already completed before.
    """

    # Nothing to do if already explored
    global ALREADY_EXPLORED
    if ALREADY_EXPLORED and not again:
        return

    _explore_packages(logger)
    _explore_directories(logger)
    _explore_entry_points(logger)

    ALREADY_EXPLORED = True


# ignore[type-var,misc]: the actual type is provided by caller - the
# return value would be assigned a name, with a narrower module type.
# This type would be propagated into this type var.
def _import(
    *,
    module: str,
    logger: Logger,
) -> ModuleT:  # type: ignore[type-var,misc]
    """
    Import a module.

    :param module: name of a module to import. It may represent a
        submodule as well, using common dot notation (``foo.bar.baz``).
    :returns: imported module.
    :raises tmt.utils.GeneralError: when import fails.
    """

    if module in sys.modules:
        logger.debug(f"Module '{module}' already imported.")

        return cast(ModuleT, sys.modules[module])

    try:
        imported = cast(ModuleT, importlib.import_module(module))

    except ImportError as exc:
        raise tmt.utils.GeneralError("Failed to import the '{module}' module.") from exc

    if module not in sys.modules:
        raise tmt.utils.GeneralError(f"Module '{module}' imported but not accessible.")

    logger.debug(f"Successfully imported the '{module}' module.")

    return imported


# ignore[type-var,misc]: the actual type is provided by caller - the
# return value would be assigned a name, with a narrower module type.
# This type would be propagated into this type var.
def _import_or_raise(
    *,
    module: str,
    exc_class: type[BaseException],
    exc_message: Optional[str] = None,
    hint_id: Optional[str] = None,
    logger: Logger,
) -> ModuleT:  # type: ignore[type-var,misc]
    """
    Import a module, or raise an exception.

    :param module: name of a module to import. It may represent a submodule as well,
        using common dot notation (``foo.bar.baz``).
    :param exc_class: an exception class to raise on failure.
    :param exc_message: an exception message.
    :returns: imported module.
    """

    try:
        return _import(module=module, logger=logger)

    except tmt.utils.GeneralError as exc:
        if hint_id is not None:
            from tmt.utils.hints import print_hints

            print_hints(hint_id, logger=logger)

        if exc_message is not None:
            raise exc_class(exc_message) from exc

        raise exc_class(f"Failed to import the '{module}' module.") from exc


# ignore[type-var,misc]: the actual type is provided by caller - the
# return value would be assigned a name, with a narrower module type.
# This type would be propagated into this type var.
def import_module(
    *,
    module: str,
    path: Optional[Path] = None,
    logger: Logger,
) -> ModuleT:  # type: ignore[type-var,misc]
    """
    Import a module.

    :param module: name of a module to import. It may represent a
        submodule as well, using common dot notation (``foo.bar.baz``).
    :param path: if specified, it would be incorporated in exception
        message.
    :returns: imported module.
    :raises SystemExit: when import fails.
    """

    path = path or Path.cwd()

    return _import_or_raise(
        module=module,
        exc_class=SystemExit,
        exc_message=f"Failed to import the '{module}' module from '{path}'.",
        logger=logger,
    )


def import_member(
    *,
    module: str,
    member: str,
    logger: Logger,
) -> tuple[ModuleT, Any]:
    """
    Import member from given module, handle errors nicely
    """

    imported: ModuleT = import_module(module=module, logger=logger)

    # Get the member and return it
    if not hasattr(imported, member):
        raise tmt.utils.GeneralError(f"No such member '{member}' in module '{module}'.")

    return (imported, getattr(imported, member))


RegisterableT = TypeVar('RegisterableT')


class PluginRegistry(Generic[RegisterableT]):
    """
    A container for plugins of shared purpose.

    A fancy wrapper for a dictionary at its core, but allows for nicer
    annotations and more visible semantics.
    """

    _plugins: dict[str, RegisterableT]

    def __init__(self, name: str) -> None:
        self.name = name

        self._plugins = {}

    def register_plugin(
        self,
        *,
        plugin_id: str,
        plugin: RegisterableT,
        raise_on_conflict: bool = True,
        logger: Logger,
    ) -> None:
        """
        Register a plugin with this registry.

        :param plugin_id: id of the plugin. Works as a label or name, and must
            may not be used in this registry yet.
        :param plugin: a plugin to register.
        :param raise_on_conflict: if set, an exception would be raised when
            id was already used.

            .. note::

               As of now, only a warning is logged, no exception is raised.
               Plugin discovery often happens in import time, and it is hard
               to manage it correctly without more changes in code.
        :param logger: used for logging.
        """

        if plugin_id in self._plugins and raise_on_conflict:
            # TODO: would be raising an exception better? Probably, but since
            # plugin discovery happens in import time, it's very hard to manage
            # it. For now, report a warning, but do not raise an exception yet.
            logger.warning(
                f"Registering plugin '{plugin}' collides"
                f" with an already registered id '{plugin_id}'"
                f" of plugin '{self._plugins[plugin_id]}'."
            )

            # raise tmt.utils.GeneralError(
            #     f"Registering plugin '{plugin.__module__}' collides"
            #     f" with an already registered id '{plugin_id}'"
            #     f" of plugin '{self._plugins[plugin_id]}'.")

        self._plugins[plugin_id] = plugin

        logger.debug(f"Registered plugin '{plugin}' with id '{plugin_id}'.")

    def get_plugin(self, plugin_id: str) -> Optional[RegisterableT]:
        """
        Find a plugin by its id.

        :returns: plugin or ``None`` if no such id has been registered.
        """

        return self._plugins.get(plugin_id, None)

    def iter_plugin_ids(self) -> Iterator[str]:
        yield from self._plugins.keys()

    def iter_plugins(self) -> Iterator[RegisterableT]:
        yield from self._plugins.values()

    def items(self) -> Iterator[tuple[str, RegisterableT]]:
        yield from self._plugins.items()

    def __len__(self) -> int:
        return len(self._plugins)

    def __bool__(self) -> bool:
        return bool(self._plugins)


class ModuleImporter(Generic[ModuleT]):
    """
    Import and return a module when called.

    A helper class for importing modules that cannot be imported in boot
    time. Constructs a callable that, when called, would import and
    return a given module. The module may be already imported, then it's
    taken from :py:attr:`sys.modules`.
    """

    def __init__(self, module: str, exc_class: type[Exception], hint_id: str) -> None:
        self._module_name = module
        self._exc_class = exc_class
        self._hint_id = hint_id

        self._module: Optional[ModuleT] = None

    def __call__(self, logger: Logger) -> ModuleT:
        if self._module is None:
            self._module = _import_or_raise(
                module=self._module_name,
                exc_class=self._exc_class,
                hint_id=self._hint_id,
                logger=logger,
            )

        assert self._module  # narrow type
        return self._module


def iter_plugin_registries() -> Iterator[PluginRegistry[Any]]:
    # TODO: maybe there is a better way, but as of now, registries do
    # not report their existence, there is no method to iterate over
    # them. Using a static list for now.
    from tmt.base import Plan, Story, Test
    from tmt.checks import _CHECK_PLUGIN_REGISTRY
    from tmt.frameworks import _FRAMEWORK_PLUGIN_REGISTRY
    from tmt.package_managers import _PACKAGE_MANAGER_PLUGIN_REGISTRY
    from tmt.plugins.plan_shapers import _PLAN_SHAPER_PLUGIN_REGISTRY
    from tmt.steps.discover import DiscoverPlugin
    from tmt.steps.execute import ExecutePlugin
    from tmt.steps.finish import FinishPlugin
    from tmt.steps.prepare import PreparePlugin
    from tmt.steps.prepare.feature import _FEATURE_PLUGIN_REGISTRY
    from tmt.steps.provision import ProvisionPlugin
    from tmt.steps.report import ReportPlugin

    yield Story._export_plugin_registry
    yield Plan._export_plugin_registry
    yield Test._export_plugin_registry
    yield _CHECK_PLUGIN_REGISTRY
    yield _FRAMEWORK_PLUGIN_REGISTRY
    yield _PLAN_SHAPER_PLUGIN_REGISTRY
    yield _PACKAGE_MANAGER_PLUGIN_REGISTRY
    yield _FEATURE_PLUGIN_REGISTRY
    yield DiscoverPlugin._supported_methods
    yield ProvisionPlugin._supported_methods
    yield PreparePlugin._supported_methods
    yield ExecutePlugin._supported_methods
    yield FinishPlugin._supported_methods
    yield ReportPlugin._supported_methods


def iter_plugins() -> Iterator[tuple[PluginRegistry[RegisterableT], str, RegisterableT]]:
    for registry in iter_plugin_registries():
        for plugin_id, plugin_class in registry.items():
            yield registry, plugin_id, plugin_class



import pathlib
from typing import Any

import _pytest.logging
import _pytest.tmpdir
import fmf
import py.path
import pytest
from pytest_container.container import ContainerData

from tmt.log import Logger
from tmt.steps.provision.podman import GuestContainer, PodmanGuestData
from tmt.utils import Path


@pytest.fixture(name='root_logger')
def fixture_root_logger(caplog: _pytest.logging.LogCaptureFixture) -> Logger:
    return Logger.create(verbose=0, debug=0, quiet=False, apply_colors_logging=False)


# Temporary directories and paths
#
# * the recommended way is to use `tmp_path` and `tmp_path_factory` fixtures
# * `tmp_path*` fixtures are not available in RHEL-8, `tmpdir` (and `tmpdir_factory`)
#   fixtures are - but these return `py.path.local` instead of a lovely `pathlib.Path`
# * `pathlib.Path` is also not good enough, as it may lack some methods in older
#   Python versions, that's why we have our own `tmt.utils.Path`.
#
# So, what we need:
#
# * a single name, we can't switch between `tmp_path` and `tmpdir` in every test
# * `tmt.utils.Path`, no strings, no `py.path.local`
# * works across all supported Python versions, from 3.6 in RHEL8 till 3.11 or so
#
# To solve this, we add here:
#
# * a wrapper class, representing the "tmp path factory". It's initialized with an
#   actual factory, has the same public API, but returns our `tmt.utils.Path`
# * two new fixtures, `tmppath` and `tmppath_factory` that consume available fixtures
#   and return corresponding `tmt.utils.Path`.
# * tests using `tmppath*` instead of pytest's own `tmp_path*` and `tmpdir*` fixtures
#
class TempPathFactory:
    def __init__(self, actual_factory: Any) -> None:
        self._actual_factory = actual_factory

    def getbasetemp(self) -> Path:
        return Path(str(self._actual_factory.getbasetemp()))

    def mktemp(self, basename: str, numbered: bool = True) -> Path:
        return Path(str(self._actual_factory.mktemp(basename, numbered=numbered)))


try:
    # If the import succeeds, we're about to wrap `Path` by `tmp_path`...
    from _pytest.tmpdir import tmp_path_factory  # noqa: F401

    @pytest.fixture(scope='session')
    def tmppath_factory(tmp_path_factory: '_pytest.tmpdir.TempPathFactory') -> TempPathFactory:  # noqa: F811
        return TempPathFactory(tmp_path_factory)

    @pytest.fixture
    def tmppath(tmp_path: pathlib.Path) -> Path:  # noqa: TID251
        return Path(str(tmp_path))

except ImportError:
    # ... and if the import fails, we're wrapping `py.path.local` from `tmpdir` family.

    # ignore[name-defined]: when inspected with our daily Python 3.9 or something,
    # the pytest is probably way newer than the one in RHEL8, and therefore the
    # name indeed would not exist. But this whole branch applies only with the old
    # pytest, therefore things are safe.
    @pytest.fixture(scope='session')
    def tmppath_factory(
        tmpdir_factory: '_pytest.tmpdir.TempdirFactory',  # type: ignore[name-defined]
    ) -> TempPathFactory:
        return TempPathFactory(tmpdir_factory)

    @pytest.fixture
    def tmppath(tmpdir: py.path.local) -> Path:
        return Path(str(tmpdir))


@pytest.fixture(scope='module')
def source_dir(tmppath_factory: TempPathFactory) -> Path:
    """
    Create dummy directory structure and remove it after tests
    """

    source_location = tmppath_factory.mktemp('source')
    (source_location / 'library').mkdir(parents=True)
    (source_location / 'lib_folder').mkdir()
    (source_location / 'tests').mkdir()
    for num in range(10):
        test_path = source_location / f'tests/bz{num}'
        test_path.mkdir()
        (test_path / 'runtests.sh').touch()
    return source_location


@pytest.fixture
def target_dir(tmppath_factory: TempPathFactory) -> Path:
    """
    Return target directory path and clean up after tests
    """

    return tmppath_factory.mktemp('target')


# Present two trees we have for identifier unit tests as fixtures, to make them
# usable in other tests as well.
@pytest.fixture(name='id_tree_defined')
def fixture_id_tree_defined() -> fmf.Tree:
    return fmf.Tree(Path(__file__).parent / 'id' / 'defined')


@pytest.fixture(name='id_tree_empty')
def fixture_id_tree_empty() -> fmf.Tree:
    return fmf.Tree(Path(__file__).parent / 'id' / 'empty')


@pytest.fixture(name='guest')
def fixture_guest(container: ContainerData, root_logger: Logger) -> GuestContainer:
    guest_data = PodmanGuestData(image=container.image_url_or_id, container=container.container_id)

    guest = GuestContainer(logger=root_logger, data=guest_data, name='dummy-container')

    guest.start()

    return guest


@pytest.fixture(name='guest_per_test')
def fixture_guest_per_test(
    container_per_test: ContainerData, root_logger: Logger
) -> GuestContainer:
    guest_data = PodmanGuestData(
        image=container_per_test.image_url_or_id, container=container_per_test.container_id
    )

    guest = GuestContainer(logger=root_logger, data=guest_data, name='dummy-container')

    guest.start()

    return guest



import pytest

from tmt.convert import relevancy_to_adjust
from tmt.utils import ConvertError


@pytest.fixture
def mini(root_logger):
    """
    Minimal example
    """

    return relevancy_to_adjust("distro = fedora: False", root_logger)


@pytest.fixture
def full(root_logger):
    """
    Full example
    """

    return relevancy_to_adjust(
        """
    # feature has been added in Fedora 33
    distro < fedora-33: False

    # using logical operators
    component = firefox && arch = ppc64: False

    arch = s390x: PHASES=novalgrind # modify environment

    # try special operators
    collection contains httpd24 && fips defined: False
    """.replace('    ', ''),
        root_logger,
    )


def check(condition, expected, logger):
    """
    Check condition against expected
    """

    adjusted = relevancy_to_adjust(f"{condition}: False", logger)[0]['when']
    assert adjusted == expected


# Valid rules


def test_empty(root_logger):
    """
    Empty relevancy
    """

    assert relevancy_to_adjust('', root_logger) == []


def test_comments(full):
    """
    Extract comments
    """

    assert full[0]['because'] == 'feature has been added in Fedora 33'
    assert full[1]['because'] == 'using logical operators'
    assert full[2]['because'] == 'modify environment'


def test_disable(mini, full):
    """
    Disable test
    """

    assert mini[0]['enabled'] is False
    assert full[0]['enabled'] is False
    assert full[1]['enabled'] is False


def test_environment(full):
    """
    Modify environment
    """

    assert full[2]['environment'] == {'PHASES': 'novalgrind'}


def test_continue(mini):
    """
    Explicit continue
    """

    assert mini[0]['continue'] is False


def test_condition(mini, full):
    """
    Expressions conversion
    """

    assert mini[0]['when'] == 'distro == fedora'
    assert full[0]['when'] == 'distro < fedora-33'
    assert full[1]['when'] == 'component == firefox and arch == ppc64'
    assert full[2]['when'] == 'arch == s390x'
    assert full[3]['when'] == 'collection == httpd24 and fips is defined'


def test_operators_basic(root_logger):
    """
    Basic operators unchanged
    """

    check('component = python', 'component == python', root_logger)
    check('component == python', 'component == python', root_logger)
    check('arch == s390x', 'arch == s390x', root_logger)
    check('arch != s390x', 'arch != s390x', root_logger)


def test_operators_distro_name(root_logger):
    """
    Check distro name
    """

    check('distro = fedora', 'distro == fedora', root_logger)
    check('distro == fedora', 'distro == fedora', root_logger)
    check('distro != fedora', 'distro != fedora', root_logger)


def test_operators_distro_major(root_logger):
    """
    Check distro major version
    """

    check('distro < fedora-33', 'distro < fedora-33', root_logger)
    check('distro > fedora-33', 'distro > fedora-33', root_logger)
    check('distro <= fedora-33', 'distro <= fedora-33', root_logger)
    check('distro >= fedora-33', 'distro >= fedora-33', root_logger)


def test_operators_distro_minor(root_logger):
    """
    Check distro minor version
    """

    check('distro = centos-8.3', 'distro ~= centos-8.3', root_logger)
    check('distro == centos-8.3', 'distro ~= centos-8.3', root_logger)
    check('distro != centos-8.3', 'distro ~!= centos-8.3', root_logger)
    check('distro < centos-8.3', 'distro ~< centos-8.3', root_logger)
    check('distro > centos-8.3', 'distro ~> centos-8.3', root_logger)
    check('distro <= centos-8.3', 'distro ~<= centos-8.3', root_logger)
    check('distro >= centos-8.3', 'distro ~>= centos-8.3', root_logger)


def test_operators_product(root_logger):
    """
    Special handling for product
    """

    # rhscl
    check('product = rhscl', 'product == rhscl', root_logger)
    check('product == rhscl', 'product == rhscl', root_logger)
    check('product != rhscl', 'product != rhscl', root_logger)
    # rhscl-3
    check('product < rhscl-3', 'product < rhscl-3', root_logger)
    check('product > rhscl-3', 'product > rhscl-3', root_logger)
    check('product <= rhscl-3', 'product <= rhscl-3', root_logger)
    check('product >= rhscl-3', 'product >= rhscl-3', root_logger)
    # rhscl-3.3
    check('product < rhscl-3.3', 'product ~< rhscl-3.3', root_logger)
    check('product > rhscl-3.3', 'product ~> rhscl-3.3', root_logger)
    check('product <= rhscl-3.3', 'product ~<= rhscl-3.3', root_logger)
    check('product >= rhscl-3.3', 'product ~>= rhscl-3.3', root_logger)


def test_operators_special(root_logger):
    """
    Check 'defined' and 'contains'
    """

    check('fips defined', 'fips is defined', root_logger)
    check('fips !defined', 'fips is not defined', root_logger)
    check('collection contains http24', 'collection == http24', root_logger)
    check('collection !contains http24', 'collection != http24', root_logger)


def test_not_equal_comma_separated(root_logger):
    """
    Special handling for comma-separated values with !=
    """

    check('distro != centos-7, centos-8', 'distro != centos-7 and distro != centos-8', root_logger)


# Invalid rules


def test_invalid_rule(root_logger):
    """
    Invalid relevancy rule
    """

    with pytest.raises(ConvertError, match='Invalid.*rule'):
        relevancy_to_adjust("weird", root_logger)


def test_invalid_decision(root_logger):
    """
    Invalid relevancy decision
    """

    with pytest.raises(ConvertError, match='Invalid.*decision'):
        relevancy_to_adjust("distro < fedora-33: weird", root_logger)


def test_invalid_expression(root_logger):
    """
    Invalid relevancy expression
    """

    with pytest.raises(ConvertError, match='Invalid.*expression'):
        relevancy_to_adjust("distro * fedora-33: False", root_logger)


def test_invalid_operator(root_logger):
    """
    Invalid relevancy operator
    """

    with pytest.raises(ConvertError, match='Invalid.*operator'):
        relevancy_to_adjust("distro <> fedora-33: False", root_logger)
